#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

from spetlr.configurator.sql.sqlparse import sql
from spetlr.configurator.sql.sqlparse import tokens as T


class StatementSplitter:
    """Filter that split stream at individual statements"""

    def __init__(self):
        self._reset()

    def _reset(self):
        """Set the filter attributes to its default values"""
        self._in_declare = False
        self._is_create = False
        self._begin_depth = 0

        self.consume_ws = False
        self.tokens = []
        self.level = 0

    def _change_splitlevel(self, ttype, value):
        """Get the new split level (increase, decrease or remain equal)"""

        # parenthesis increase/decrease a level
        if ttype is T.Punctuation and value == "(":
            return 1
        elif ttype is T.Punctuation and value == ")":
            return -1
        elif ttype not in T.Keyword:  # if normal token return
            return 0

        # Everything after here is ttype = T.Keyword
        # Also to note, once entered an If statement you are done and basically
        # returning
        unified = value.upper()

        # three keywords begin with CREATE, but only one of them is DDL
        # DDL Create though can contain more words such as "or replace"
        if ttype is T.Keyword.DDL and unified.startswith("CREATE"):
            self._is_create = True
            return 0

        # can have nested declare inside of being...
        if unified == "DECLARE" and self._is_create and self._begin_depth == 0:
            self._in_declare = True
            return 1

        if unified == "BEGIN":
            self._begin_depth += 1
            if self._is_create:
                # FIXME(andi): This makes no sense.
                return 1
            return 0

        # Should this respect a preceding BEGIN?
        # In CASE ... WHEN ... END this results in a split level -1.
        # Would having multiple CASE WHEN END and a Assignment Operator
        # cause the statement to cut off prematurely?
        if unified == "END":
            self._begin_depth = max(0, self._begin_depth - 1)
            return -1

        if (
            unified in ("IF", "FOR", "WHILE", "CASE")
            and self._is_create
            and self._begin_depth > 0
        ):
            return 1

        if unified in ("END IF", "END FOR", "END WHILE"):
            return -1

        # Default
        return 0

    def process(self, stream):
        """Process the stream"""
        EOS_TTYPE = T.Whitespace, T.Comment.Single

        # Run over all stream tokens
        for ttype, value in stream:
            # Yield token if we finished a statement and there's no whitespaces
            # It will count newline token as a non whitespace. In this context
            # whitespace ignores newlines.
            # why don't multi line comments also count?
            if self.consume_ws and ttype not in EOS_TTYPE:
                yield sql.Statement(self.tokens)

                # Reset filter and prepare to process next statement
                self._reset()

            # Change current split level (increase, decrease or remain equal)
            self.level += self._change_splitlevel(ttype, value)

            # Append the token to the current statement
            self.tokens.append(sql.Token(ttype, value))

            # Check if we get the end of a statement
            if self.level <= 0 and ttype is T.Punctuation and value == ";":
                self.consume_ws = True

        # Yield pending statement (if any)
        if self.tokens and not all(t.is_whitespace for t in self.tokens):
            yield sql.Statement(self.tokens)



#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

from spetlr.configurator.sql.sqlparse.engine import grouping
from spetlr.configurator.sql.sqlparse.engine.filter_stack import FilterStack
from spetlr.configurator.sql.sqlparse.engine.statement_splitter import StatementSplitter

__all__ = [
    "grouping",
    "FilterStack",
    "StatementSplitter",
]



import unittest
from textwrap import dedent

from pyspark.sql import types as t

from spetlr import Configurator
from spetlr.schema_manager import SchemaManager
from spetlr.schema_manager.spark_schema import get_schema
from spetlr.sql import SqlExecutor
from tests.local.configurator import sql, tables1, tables2, tables3, tables4, tables5


class TestConfigurator(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        Configurator().clear_all_configurations()

    def test_01_import_config(self):
        tc = Configurator()
        tc.add_resource_path(tables1)

        tc.reset(debug=True)
        self.assertRegex(tc.table_name("MyFirst"), "first__.*")
        self.assertRegex(tc.table_name("MyAlias"), "first__.*")
        self.assertRegex(tc.table_name("MyForked"), "another")

        tc.reset(debug=False)
        self.assertRegex(tc.table_name("MyFirst"), "first")
        self.assertRegex(tc.table_name("MyAlias"), "first")
        self.assertRegex(tc.table_name("MyForked"), "first")

    def test_02_details(self):
        tc = Configurator()

        tc.reset(debug=True)
        details = tc.get_all_details()
        self.assertRegex(details["MyFirst_path"], "/tmp/my/first__.*")

        tc.reset(debug=False)
        details = tc.get_all_details()
        self.assertRegex(details["MyFirst_path"], "/mnt/my/first")

    def test_03_json(self):
        tc = Configurator()

        with self.assertRaises(KeyError):
            tc.add_resource_path(tables2)

        tc.set_extra(ENV="dev")
        tc.add_resource_path(tables2)
        tc.reset(debug=True)
        self.assertRegex(tc.table_name("MyThird"), "first__.*")

        self.assertRegex(
            tc.get_all_details()["MyFourth_path"], "/tmp/dev/path/to/delta"
        )

    def test_04_recursing(self):
        tc = Configurator()
        tc.set_prod()
        self.assertEqual(tc.get("MyRecursing", "path"), "/mnt/tables/recursing")

    def test_05_test_self_recursion_detection(self):
        tc = Configurator()
        tc.set_prod()
        with self.assertRaises(ValueError):
            tc.add_resource_path(tables3)

    def test_06_freestyle(self):
        tc = Configurator()
        tc.set_prod()
        tc.add_resource_path(tables4)
        details = tc.get_all_details()
        self.assertTrue("MyFreeTable_eggs", details)
        self.assertEqual(tc.table_property("MyFreeTable", "bacon"), "")

    def test_07_bare_strings_and_structures(self):
        tc = Configurator()
        tc.set_prod()
        tc.add_resource_path(tables5)
        self.assertEqual(tc.get("MyPlainLiteral"), "Bar")
        self.assertEqual(tc.get_all_details()["MyCompositeLiteral"], "FooBar")
        st = SchemaManager().get_schema("MyComposite")
        self.assertEqual(
            st,
            t.StructType(
                [
                    t.StructField("a", t.IntegerType()),
                    t.StructField("b", t.DecimalType(12, 2)),
                ]
            ),
        )
        self.assertEqual(tc.table_name("MyComposite"), "ottoBar")

    def test_08_test_deprecated_import(self):
        from spetlr import Configurator
        from spetlr.config_master import TableConfigurator

        tc = TableConfigurator()
        self.assertEqual(tc.table_name("MyComposite"), "ottoBar")

        c = Configurator()

        self.assertIs(tc, c)

    def test_09_configure_from_sql(self):
        c = Configurator()
        c.clear_all_configurations()
        c.add_sql_resource_path(sql)
        c.set_prod()

        self.assertEqual(c.get("MySparkDb", "name"), "my_db1")
        self.assertEqual(c.get("MySparkDb", "path"), "/tmp/foo/bar/my_db1/")
        self.assertEqual(c.get("MySparkDb", "format"), "db")

        self.assertEqual(c.get("MySqlTable", "name"), "my_db1.tbl1")
        c.set_debug()
        self.assertEqual(c.get("MySqlTable", "path"), "/tmp/foo/bar/my_db1/tbl1/")

        self.assertEqual(c.get("MySqlTable", "format"), "delta")
        self.assertEqual(
            c.get("MySqlTable", "options"), {"key1": "val1", "key2": "val2"}
        )
        self.assertEqual(c.get("MySqlTable", "partitioned_by"), ["a", "b"])
        self.assertEqual(
            c.get("MySqlTable", "clustered_by"),
            dict(
                cols=["c", "d"],
                sorted_by=[
                    dict(name="a", ordering="ASC"),
                    dict(name="b", ordering="DESC"),
                ],
                buckets=5,
            ),
        )
        self.assertEqual(c.get("MySqlTable", "comment"), "Dummy Database 1 table 1")
        self.assertEqual(
            c.get("MySqlTable", "tblproperties"),
            {"key1": "val1", "key2": "val2", "my.key.3": "true"},
        )
        self.assertEqual(
            SchemaManager().get_schema("MySqlTable"),
            get_schema("""a int, b int, c string, d timestamp"""),
        )

        self.assertEqual(
            SchemaManager().get_schema_as_string("MyDetailsTable"),
            """a int, b int, c string, d timestamp, another int""",
        )

        c.set_prod()
        statements = list(SqlExecutor(sql).get_statements("*"))
        self.assertEqual(len(statements), 3)
        self.assertEqual(
            statements[1],
            dedent(
                """\


                        -- spetlr.Configurator key: MyDetailsTable
                        CREATE TABLE IF NOT EXISTS my_db1.details
                        (
                          a int, b int, c string, d timestamp,
                          another int
                          -- comment with ;
                        )
                        USING DELTA
                        COMMENT "Dummy Database 1 details"
                        LOCATION "/mnt/foo/bar/my_db1/details/";"""
            ),
        )



from typing import Generator, List, Tuple

from spetlr.exceptions.configurator_exceptions import (
    SpetlrConfiguratorInvalidSqlException,
)

from . import sqlparse
from .StatementBlocks import ENUM_DESC, SortingCol
from .types import _PeekableTokenList
from .utils import _meaningful_token_iter


def _unpack_comma_separated_list_in_parens(
    stmt: _PeekableTokenList,
) -> Generator[List[sqlparse.tokens._TokenType], None, None]:
    grouped_token = next(stmt)
    if not isinstance(grouped_token, sqlparse.sql.Parenthesis):
        SpetlrConfiguratorInvalidSqlException("expected parenthesis with list")

    it = _meaningful_token_iter(grouped_token.flatten())
    for token in it:
        if token.value in "()":
            continue
        between_commas = [token]
        for token in it:
            if token.value == "," or token.value in "()":
                break
            between_commas.append(token)
        yield between_commas


def _unpack_list_of_single_variables(stmt: _PeekableTokenList):
    for tokens in _unpack_comma_separated_list_in_parens(stmt):
        if len(tokens) != 1:
            SpetlrConfiguratorInvalidSqlException(
                "unexpected number of statements between commas"
            )
        yield tokens[0]


def _unpack_list_of_sorting_statements(stmt: _PeekableTokenList):
    for tokens in _unpack_comma_separated_list_in_parens(stmt):
        if len(tokens) not in [1, 2]:
            SpetlrConfiguratorInvalidSqlException(
                "unexpected number of statements between commas"
            )

        sort_col = SortingCol(tokens[0].value)
        if len(tokens) == 2:
            direction = tokens[1]
            if direction.value.upper() == "ASC":
                pass
            elif direction.value.upper() == "DESC":
                sort_col.ordering = ENUM_DESC
            else:
                SpetlrConfiguratorInvalidSqlException("unexpected ordering direction")
        yield sort_col


def _unpack_options(
    stmt: _PeekableTokenList,
) -> Generator[Tuple[str, str], None, None]:
    for tokens in _unpack_comma_separated_list_in_parens(stmt):
        statement = str(sqlparse.sql.TokenList(tokens))
        assignment = statement.split("=")

        if len(assignment) != 2:
            raise SpetlrConfiguratorInvalidSqlException(
                f"expected assignments, got {statement}"
            )

        key, value = assignment

        yield key.strip("\"'"), value.strip("\"'")



from typing import List, Union

from pyspark.sql import DataFrame

from spetlr.etl import Loader, dataset_group
from spetlr.tables import TableHandle


class UpsertLoader(Loader):
    def __init__(
        self,
        handle: TableHandle,
        join_cols: List[str],
        dataset_input_keys: Union[str, List[str]] = None,
    ):
        super().__init__(dataset_input_keys=dataset_input_keys)

        self.handle = handle
        self.join_cols = join_cols

    def save_many(self, datasets: dataset_group) -> None:
        raise NotImplementedError()

    def save(self, df: DataFrame) -> None:
        """Upserts a single dataframe to the target table."""
        self.handle.upsert(df=df, join_cols=self.join_cols)



import unittest

from azure.cosmos.exceptions import CosmosHttpResponseError

import spetlr.cosmos
from spetlr import Configurator
from spetlr.functions import init_dbutils
from spetlr.spark import Spark


class TestCosmos(spetlr.cosmos.CosmosDb):
    def __init__(self):
        dbutils = init_dbutils()
        super().__init__(
            endpoint=dbutils.secrets.get("values", "Cosmos--Endpoint"),
            account_key=dbutils.secrets.get("secrets", "Cosmos--AccountKey"),
            database="SpetlrCosmosContainer",
            catalog_name=None,
        )


class CosmosTests(unittest.TestCase):
    def test_01_tables(self):
        tc = Configurator()
        tc.clear_all_configurations()
        tc.register(
            "CmsTbl",
            {
                "name": "CosmosTable",
                "schema": {"sql": "id string, pk string, value int"},
                "rows_per_partition": 5,
            },
        )

    def test_02_create_db(self):
        cm = TestCosmos()
        # clean up the database in case it exists from last run
        try:
            cm.client.delete_database(cm.database)
        except CosmosHttpResponseError:
            pass

        cm.execute_sql(
            f"CREATE DATABASE IF NOT EXISTS {cm.catalog_name}.{cm.database};"
        )
        tc = Configurator()
        cm.execute_sql(
            "CREATE TABLE IF NOT EXISTS"
            f" {cm.catalog_name}.{cm.database}.{tc.table_name('CmsTbl')}"
            " using cosmos.oltp"
            " TBLPROPERTIES(partitionKeyPath = '/pk', manualThroughput = '400')"
        )

    def test_03_write_table(self):
        cm = TestCosmos()
        df = Spark.get().createDataFrame(
            [("first", "pk1", 56), ("second", "pk2", 987)],
            "id string, pk string, value int",
        )
        cm.write_table(df, "CmsTbl")

    def test_04_read_table(self):
        cm = TestCosmos()
        df = cm.read_table("CmsTbl")
        data = set(tuple(row) for row in df.collect())
        self.assertEqual({("first", "pk1", 56), ("second", "pk2", 987)}, data)

    def test_05_use_handle(self):
        ch = TestCosmos().from_tc("CmsTbl")
        df = ch.read()
        data = set(tuple(row) for row in df.collect())
        self.assertEqual({("first", "pk1", 56), ("second", "pk2", 987)}, data)

        # create a df with two new rows.
        new_df = Spark.get().createDataFrame(
            [("third", "pk1", 56), ("fourth", "pk2", 987)],
            "id string, pk string, value int",
        )

        # in append mode, there will now be 4 rows.
        ch.append(new_df)
        self.assertEqual(ch.read().count(), 4)

        # drop and recreate the container
        # we should end up with only the new rows.
        ch.recreate()
        ch.append(new_df)
        self.assertEqual(ch.read().count(), 2)

    def test_10_delete_items(self):
        cm = TestCosmos()
        cm.delete_item("CmsTbl", "third", "pk1")
        cm.delete_item("CmsTbl", "fourth", "pk2")

        cm.client.delete_database(cm.database)



#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

from spetlr.configurator.sql.sqlparse.filters.aligned_indent import AlignedIndentFilter
from spetlr.configurator.sql.sqlparse.filters.others import (
    SerializerUnicode,
    SpacesAroundOperatorsFilter,
    StripCommentsFilter,
    StripWhitespaceFilter,
)
from spetlr.configurator.sql.sqlparse.filters.output import (
    OutputPHPFilter,
    OutputPythonFilter,
)
from spetlr.configurator.sql.sqlparse.filters.reindent import ReindentFilter
from spetlr.configurator.sql.sqlparse.filters.right_margin import RightMarginFilter
from spetlr.configurator.sql.sqlparse.filters.tokens import (
    IdentifierCaseFilter,
    KeywordCaseFilter,
    TruncateStringFilter,
)

__all__ = [
    "SerializerUnicode",
    "StripCommentsFilter",
    "StripWhitespaceFilter",
    "SpacesAroundOperatorsFilter",
    "OutputPHPFilter",
    "OutputPythonFilter",
    "KeywordCaseFilter",
    "IdentifierCaseFilter",
    "TruncateStringFilter",
    "ReindentFilter",
    "RightMarginFilter",
    "AlignedIndentFilter",
]



from abc import abstractmethod
from typing import List, Union

from pyspark.sql import DataFrame

from spetlr.etl.types import EtlBase, dataset_group


class TransformerNC(EtlBase):
    """If you only want to transform a single input dataframe,
    implement `process`
    If you want to transform a set of dataframes,
    implement `process_many`

    In regards to the etl step, the TransformerNC does NOT CONSUME the inputs
    and ADDs the result of its transformation stage to the dataset dict.
    """

    def __init__(
        self,
        *,
        dataset_input_keys: Union[str, List[str]] = None,
        dataset_output_key: str = None,
    ):
        if dataset_output_key is not None:
            self.dataset_output_key = dataset_output_key
        else:
            self.dataset_output_key = type(self).__name__

        if dataset_input_keys is None:
            self.dataset_input_key_list = []
        elif isinstance(dataset_input_keys, str):
            self.dataset_input_key_list = [dataset_input_keys]
        else:
            self.dataset_input_key_list = dataset_input_keys

    def etl(self, inputs: dataset_group) -> dataset_group:
        if len(self.dataset_input_key_list) > 0:
            if len(self.dataset_input_key_list) == 1:
                df = self.process(inputs[self.dataset_input_key_list[0]])
            else:
                datasetFilteret = {
                    datasetKey: df
                    for datasetKey, df in inputs.items()
                    if datasetKey in self.dataset_input_key_list
                }
                df = self.process_many(datasetFilteret)
        elif len(inputs) == 1:
            df = self.process(next(iter(inputs.values())))
        else:
            df = self.process_many(inputs)

        inputs[self.dataset_output_key] = df
        return inputs

    @abstractmethod
    def process(self, df: DataFrame) -> DataFrame:
        raise NotImplementedError()

    @abstractmethod
    def process_many(self, datasets: dataset_group) -> DataFrame:
        raise NotImplementedError()



import unittest

from pyspark.sql import DataFrame
from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from spetlr.etl import Transformer
from spetlr.etl.types import dataset_group
from spetlr.spark import Spark


class TransformerTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.transformer = TestTransformer()
        cls.df = create_dataframe()

    def test_process_returns_dataframe(self):
        result = self.transformer.etl({"df": self.df})
        self.assertEqual({"TestTransformer"}, set(result.keys()))
        self.assertIs(list(result.values())[0], self.df)

    def test_process_many(self):
        result = self.transformer.etl({"df1": self.df, "df2": self.df})
        self.assertEqual({"TestTransformer"}, set(result.keys()))
        self.assertEqual(6, list(result.values())[0].count())

    def test_dataset_key(self):
        result = TestTransformer("my_key").etl({"df": self.df})
        self.assertEqual({"my_key"}, set(result.keys()))


class TestTransformer(Transformer):
    def process(self, df: DataFrame) -> DataFrame:
        return df

    def process_many(self, datasets: dataset_group) -> DataFrame:
        assert len(datasets) == 2
        df1, df2 = list(datasets.values())
        return df1.union(df2)


def create_dataframe():
    data = [(1, "1"), (2, "2"), (3, "3")]

    schema = StructType(
        [
            StructField("id", IntegerType(), False),
            StructField("text", StringType(), False),
        ]
    )

    return Spark.get().createDataFrame(data=data, schema=schema)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

"""SQL Lexer"""
import re
from io import TextIOBase

from spetlr.configurator.sql.sqlparse import keywords, tokens
from spetlr.configurator.sql.sqlparse.utils import consume

# This code is based on the SqlLexer in pygments.
# http://pygments.org/
# It's separated from the rest of pygments to increase performance
# and to allow some customizations.


class Lexer:
    """The Lexer supports configurable syntax.
    To add support for additional keywords, use the `add_keywords` method."""

    _default_intance = None

    # Development notes:
    # - This class is prepared to be able to support additional SQL dialects
    #   in the future by adding additional functions that take the place of
    #   the function default_initialization()
    # - The lexer class uses an explicit singleton behavior with the
    #   instance-getter method get_default_instance(). This mechanism has
    #   the advantage that the call signature of the entry-points to the
    #   sqlparse library are not affected. Also, usage of sqlparse in third
    #   party code does not need to be adapted. On the other hand, singleton
    #   behavior is not thread safe, and the current implementation does not
    #   easily allow for multiple SQL dialects to be parsed in the same
    #   process. Such behavior can be supported in the future by passing a
    #   suitably initialized lexer object as an additional parameter to the
    #   entry-point functions (such as `parse`). Code will need to be written
    #   to pass down and utilize such an object. The current implementation
    #   is prepared to support this thread safe approach without the
    #   default_instance part needing to change interface.

    @classmethod
    def get_default_instance(cls):
        """Returns the lexer instance used internally
        by the sqlparse core functions."""
        if cls._default_intance is None:
            cls._default_intance = cls()
            cls._default_intance.default_initialization()
        return cls._default_intance

    def default_initialization(self):
        """Initialize the lexer with default dictionaries.
        Useful if you need to revert custom syntax settings."""
        self.clear()
        self.set_SQL_REGEX(keywords.SQL_REGEX)
        self.add_keywords(keywords.KEYWORDS_COMMON)
        self.add_keywords(keywords.KEYWORDS_ORACLE)
        self.add_keywords(keywords.KEYWORDS_PLPGSQL)
        self.add_keywords(keywords.KEYWORDS_HQL)
        self.add_keywords(keywords.KEYWORDS_MSACCESS)
        self.add_keywords(keywords.KEYWORDS)

    def clear(self):
        """Clear all syntax configurations.
        Useful if you want to load a reduced set of syntax configurations.
        After this call, regexps and keyword dictionaries need to be loaded
        to make the lexer functional again."""
        self._SQL_REGEX = []
        self._keywords = []

    def set_SQL_REGEX(self, SQL_REGEX):
        """Set the list of regex that will parse the SQL."""
        FLAGS = re.IGNORECASE | re.UNICODE
        self._SQL_REGEX = [(re.compile(rx, FLAGS).match, tt) for rx, tt in SQL_REGEX]
        for regex, tt in SQL_REGEX:
            if not (
                isinstance(tt, tokens._TokenType) or tt is keywords.PROCESS_AS_KEYWORD
            ):
                raise AssertionError(f"Invalid action for regex {repr(regex)}")

    def add_keywords(self, keywords):
        """Add keyword dictionaries. Keywords are looked up in the same order
        that dictionaries were added."""
        self._keywords.append(keywords)

    def is_keyword(self, value):
        """Checks for a keyword.

        If the given value is in one of the KEYWORDS_* dictionary
        it's considered a keyword. Otherwise, tokens.Name is returned.
        """
        val = value.upper()
        for kwdict in self._keywords:
            if val in kwdict:
                return kwdict[val], value
        else:
            return tokens.Name, value

    def get_tokens(self, text, encoding=None):
        """
        Return an iterable of (tokentype, value) pairs generated from
        `text`. If `unfiltered` is set to `True`, the filtering mechanism
        is bypassed even if filters are defined.

        Also preprocess the text, i.e. expand tabs and strip it if
        wanted and applies registered filters.

        Split ``text`` into (tokentype, text) pairs.

        ``stack`` is the initial stack (default: ``['root']``)
        """
        if isinstance(text, TextIOBase):
            text = text.read()

        if isinstance(text, str):
            pass
        elif isinstance(text, bytes):
            if encoding:
                text = text.decode(encoding)
            else:
                try:
                    text = text.decode("utf-8")
                except UnicodeDecodeError:
                    text = text.decode("unicode-escape")
        else:
            raise TypeError(
                "Expected text or file-like object, got {!r}".format(type(text))
            )

        iterable = enumerate(text)
        for pos, char in iterable:
            for rexmatch, action in self._SQL_REGEX:
                m = rexmatch(text, pos)

                if not m:
                    continue
                elif isinstance(action, tokens._TokenType):
                    yield action, m.group()
                elif action is keywords.PROCESS_AS_KEYWORD:
                    yield self.is_keyword(m.group())

                consume(iterable, m.end() - pos - 1)
                break
            else:
                yield tokens.Error, char


def tokenize(sql, encoding=None):
    """Tokenize sql.

    Tokenize *sql* using the :class:`Lexer` and return a 2-tuple stream
    of ``(token type, value)`` items.
    """
    return Lexer.get_default_instance().get_tokens(sql, encoding)



#!/usr/bin/env python
#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

"""Module that contains the command line app.

Why does this file exist, and why not put this in __main__?
  You might be tempted to import things from __main__ later, but that will
  cause problems: the code will get executed twice:
  - When you run `python -m sqlparse` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``sqlparse.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``sqlparse.__main__`` in ``sys.modules``.
  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import argparse
import sys
from io import TextIOWrapper

import sqlparse

from spetlr.configurator.sql.sqlparse.exceptions import SQLParseError


# TODO: Add CLI Tests
# TODO: Simplify formatter by using argparse `type` arguments
def create_parser():
    _CASE_CHOICES = ["upper", "lower", "capitalize"]

    parser = argparse.ArgumentParser(
        prog="sqlformat",
        description='Format FILE according to OPTIONS. Use "-" as FILE '
        "to read from stdin.",
        usage="%(prog)s  [OPTIONS] FILE, ...",
    )

    parser.add_argument("filename")

    parser.add_argument(
        "-o",
        "--outfile",
        dest="outfile",
        metavar="FILE",
        help="write output to FILE (defaults to stdout)",
    )

    parser.add_argument("--version", action="version", version=sqlparse.__version__)

    group = parser.add_argument_group("Formatting Options")

    group.add_argument(
        "-k",
        "--keywords",
        metavar="CHOICE",
        dest="keyword_case",
        choices=_CASE_CHOICES,
        help="change case of keywords, CHOICE is one of {}".format(
            ", ".join('"{}"'.format(x) for x in _CASE_CHOICES)
        ),
    )

    group.add_argument(
        "-i",
        "--identifiers",
        metavar="CHOICE",
        dest="identifier_case",
        choices=_CASE_CHOICES,
        help="change case of identifiers, CHOICE is one of {}".format(
            ", ".join('"{}"'.format(x) for x in _CASE_CHOICES)
        ),
    )

    group.add_argument(
        "-l",
        "--language",
        metavar="LANG",
        dest="output_format",
        choices=["python", "php"],
        help="output a snippet in programming language LANG, "
        'choices are "python", "php"',
    )

    group.add_argument(
        "--strip-comments",
        dest="strip_comments",
        action="store_true",
        default=False,
        help="remove comments",
    )

    group.add_argument(
        "-r",
        "--reindent",
        dest="reindent",
        action="store_true",
        default=False,
        help="reindent statements",
    )

    group.add_argument(
        "--indent_width",
        dest="indent_width",
        default=2,
        type=int,
        help="indentation width (defaults to 2 spaces)",
    )

    group.add_argument(
        "--indent_after_first",
        dest="indent_after_first",
        action="store_true",
        default=False,
        help="indent after first line of statement (e.g. SELECT)",
    )

    group.add_argument(
        "--indent_columns",
        dest="indent_columns",
        action="store_true",
        default=False,
        help="indent all columns by indent_width instead of keyword length",
    )

    group.add_argument(
        "-a",
        "--reindent_aligned",
        action="store_true",
        default=False,
        help="reindent statements to aligned format",
    )

    group.add_argument(
        "-s",
        "--use_space_around_operators",
        action="store_true",
        default=False,
        help="place spaces around mathematical operators",
    )

    group.add_argument(
        "--wrap_after",
        dest="wrap_after",
        default=0,
        type=int,
        help="Column after which lists should be wrapped",
    )

    group.add_argument(
        "--comma_first",
        dest="comma_first",
        default=False,
        type=bool,
        help="Insert linebreak before comma (default False)",
    )

    group.add_argument(
        "--encoding",
        dest="encoding",
        default="utf-8",
        help="Specify the input encoding (default utf-8)",
    )

    return parser


def _error(msg):
    """Print msg and optionally exit with return code exit_."""
    sys.stderr.write("[ERROR] {}\n".format(msg))
    return 1


def main(args=None):
    parser = create_parser()
    args = parser.parse_args(args)

    if args.filename == "-":  # read from stdin
        wrapper = TextIOWrapper(sys.stdin.buffer, encoding=args.encoding)
        try:
            data = wrapper.read()
        finally:
            wrapper.detach()
    else:
        try:
            with open(args.filename, encoding=args.encoding) as f:
                data = "".join(f.readlines())
        except OSError as e:
            return _error("Failed to read {}: {}".format(args.filename, e))

    close_stream = False
    if args.outfile:
        try:
            stream = open(args.outfile, "w", encoding=args.encoding)
            close_stream = True
        except OSError as e:
            return _error("Failed to open {}: {}".format(args.outfile, e))
    else:
        stream = sys.stdout

    formatter_opts = vars(args)
    try:
        formatter_opts = sqlparse.formatter.validate_options(formatter_opts)
    except SQLParseError as e:
        return _error("Invalid options: {}".format(e))

    s = sqlparse.format(data, **formatter_opts)
    stream.write(s)
    stream.flush()
    if close_stream:
        stream.close()
    return 0



"""
This test exists both in the 'local' and in the 'cluster' folder.
The tests are not completely equal.
The reason is that the slack notifier should work both in a job and outside one.
"""
import json
import threading
import unittest
from http.server import BaseHTTPRequestHandler, HTTPServer

from spetlr.reporting import SlackNotifier


class SlackNotifierTests(unittest.TestCase):
    httpd: HTTPServer
    serve_thread: threading.Thread
    hook_url: str

    class HTTPHandler(BaseHTTPRequestHandler):
        called = False
        called_data = {}

        def do_POST(self):
            SlackNotifierTests.HTTPHandler.called = True
            data_string = self.rfile.read(int(self.headers["Content-Length"]))

            self.send_response(200)
            self.end_headers()

            SlackNotifierTests.HTTPHandler.called_data = json.loads(data_string)

    @classmethod
    def setUpClass(cls) -> None:
        cls.httpd = HTTPServer(("localhost", 0), cls.HTTPHandler)
        cls.serve_thread = threading.Thread(target=cls.httpd.serve_forever)
        cls.serve_thread.start()
        cls.hook_url = f"http://localhost:{cls.httpd.server_port}"
        print(f"Serving webhook at {cls.hook_url}")

    @classmethod
    def tearDownClass(cls) -> None:
        cls.httpd.shutdown()

    def setUp(self) -> None:
        self.HTTPHandler.called = False
        self.HTTPHandler.called_data = {}

    def test_01_info_webhook(self):
        slack = SlackNotifier(self.hook_url)

        # Call the WebHookNotifier here
        slack.notify_info("my nice message")

        self.assertTrue(self.HTTPHandler.called)
        self.assertIn("text", self.HTTPHandler.called_data)
        self.assertRegex(
            self.HTTPHandler.called_data["text"],
            "message was sent from your job Testing Run",
        )
        self.assertRegex(self.HTTPHandler.called_data["text"], "test_01_info_webhook")
        self.assertRegex(self.HTTPHandler.called_data["text"], "my nice message")
        print(self.HTTPHandler.called_data["text"])

    def test_02_exc_webhook(self):
        slack = SlackNotifier(self.hook_url)

        try:
            1 / 0
        except ZeroDivisionError:
            slack.notify_exc()
            pass

        self.assertTrue(self.HTTPHandler.called)
        self.assertIn("text", self.HTTPHandler.called_data)
        self.assertRegex(
            self.HTTPHandler.called_data["text"],
            "exception has occurred in your job Testing Run",
        )
        self.assertRegex(self.HTTPHandler.called_data["text"], "test_02_exc_webhook")
        print(self.HTTPHandler.called_data["text"])



#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

"""This module contains classes representing syntactical elements of SQL."""

import re

from spetlr.configurator.sql.sqlparse import tokens as T
from spetlr.configurator.sql.sqlparse.utils import imt, remove_quotes


class NameAliasMixin:
    """Implements get_real_name and get_alias."""

    def get_real_name(self):
        """Returns the real name (object name) of this identifier."""
        # a.b
        dot_idx, _ = self.token_next_by(m=(T.Punctuation, "."))
        return self._get_first_name(dot_idx, real_name=True)

    def get_alias(self):
        """Returns the alias for this identifier or ``None``."""

        # "name AS alias"
        kw_idx, kw = self.token_next_by(m=(T.Keyword, "AS"))
        if kw is not None:
            return self._get_first_name(kw_idx + 1, keywords=True)

        # "name alias" or "complicated column expression alias"
        _, ws = self.token_next_by(t=T.Whitespace)
        if len(self.tokens) > 2 and ws is not None:
            return self._get_first_name(reverse=True)


class Token:
    """Base class for all other classes in this module.

    It represents a single token and has two instance attributes:
    ``value`` is the unchanged value of the token and ``ttype`` is
    the type of the token.
    """

    __slots__ = (
        "value",
        "ttype",
        "parent",
        "normalized",
        "is_keyword",
        "is_group",
        "is_whitespace",
    )

    def __init__(self, ttype, value):
        value = str(value)
        self.value = value
        self.ttype = ttype
        self.parent = None
        self.is_group = False
        self.is_keyword = ttype in T.Keyword
        self.is_whitespace = self.ttype in T.Whitespace
        self.normalized = value.upper() if self.is_keyword else value

    def __str__(self):
        return self.value

    # Pending tokenlist __len__ bug fix
    # def __len__(self):
    #     return len(self.value)

    def __repr__(self):
        cls = self._get_repr_name()
        value = self._get_repr_value()

        q = '"' if value.startswith("'") and value.endswith("'") else "'"
        return "<{cls} {q}{value}{q} at 0x{id:2X}>".format(id=id(self), **locals())

    def _get_repr_name(self):
        return str(self.ttype).split(".")[-1]

    def _get_repr_value(self):
        raw = str(self)
        if len(raw) > 7:
            raw = raw[:6] + "..."
        return re.sub(r"\s+", " ", raw)

    def flatten(self):
        """Resolve subgroups."""
        yield self

    def match(self, ttype, values, regex=False):
        """Checks whether the token matches the given arguments.

        *ttype* is a token type. If this token doesn't match the given token
        type.
        *values* is a list of possible values for this token. The values
        are OR'ed together so if only one of the values matches ``True``
        is returned. Except for keyword tokens the comparison is
        case-sensitive. For convenience it's OK to pass in a single string.
        If *regex* is ``True`` (default is ``False``) the given values are
        treated as regular expressions.
        """
        type_matched = self.ttype is ttype
        if not type_matched or values is None:
            return type_matched

        if isinstance(values, str):
            values = (values,)

        if regex:
            # TODO: Add test for regex with is_keyboard = false
            flag = re.IGNORECASE if self.is_keyword else 0
            values = (re.compile(v, flag) for v in values)

            for pattern in values:
                if pattern.search(self.normalized):
                    return True
            return False

        if self.is_keyword:
            values = (v.upper() for v in values)

        return self.normalized in values

    def within(self, group_cls):
        """Returns ``True`` if this token is within *group_cls*.

        Use this method for example to check if an identifier is within
        a function: ``t.within(sql.Function)``.
        """
        parent = self.parent
        while parent:
            if isinstance(parent, group_cls):
                return True
            parent = parent.parent
        return False

    def is_child_of(self, other):
        """Returns ``True`` if this token is a direct child of *other*."""
        return self.parent == other

    def has_ancestor(self, other):
        """Returns ``True`` if *other* is in this tokens ancestry."""
        parent = self.parent
        while parent:
            if parent == other:
                return True
            parent = parent.parent
        return False


class TokenList(Token):
    """A group of tokens.

    It has an additional instance attribute ``tokens`` which holds a
    list of child-tokens.
    """

    __slots__ = "tokens"

    def __init__(self, tokens=None):
        self.tokens = tokens or []
        [setattr(token, "parent", self) for token in self.tokens]
        super().__init__(None, str(self))
        self.is_group = True

    def __str__(self):
        return "".join(token.value for token in self.flatten())

    # weird bug
    # def __len__(self):
    #     return len(self.tokens)

    def __iter__(self):
        return iter(self.tokens)

    def __getitem__(self, item):
        return self.tokens[item]

    def _get_repr_name(self):
        return type(self).__name__

    def _pprint_tree(self, max_depth=None, depth=0, f=None, _pre=""):
        """Pretty-print the object tree."""
        token_count = len(self.tokens)
        for idx, token in enumerate(self.tokens):
            cls = token._get_repr_name()
            value = token._get_repr_value()

            last = idx == (token_count - 1)
            pre = "`- " if last else "|- "

            q = '"' if value.startswith("'") and value.endswith("'") else "'"
            print("{_pre}{pre}{idx} {cls} {q}{value}{q}".format(**locals()), file=f)

            if token.is_group and (max_depth is None or depth < max_depth):
                parent_pre = "   " if last else "|  "
                token._pprint_tree(max_depth, depth + 1, f, _pre + parent_pre)

    def get_token_at_offset(self, offset):
        """Returns the token that is on position offset."""
        idx = 0
        for token in self.flatten():
            end = idx + len(token.value)
            if idx <= offset < end:
                return token
            idx = end

    def flatten(self):
        """Generator yielding ungrouped tokens.

        This method is recursively called for all child tokens.
        """
        for token in self.tokens:
            if token.is_group:
                yield from token.flatten()
            else:
                yield token

    def get_sublists(self):
        for token in self.tokens:
            if token.is_group:
                yield token

    @property
    def _groupable_tokens(self):
        return self.tokens

    def _token_matching(self, funcs, start=0, end=None, reverse=False):
        """next token that match functions"""
        if start is None:
            return None

        if not isinstance(funcs, (list, tuple)):
            funcs = (funcs,)

        if reverse:
            assert end is None
            indexes = range(start - 2, -1, -1)
        else:
            if end is None:
                end = len(self.tokens)
            indexes = range(start, end)
        for idx in indexes:
            token = self.tokens[idx]
            for func in funcs:
                if func(token):
                    return idx, token
        return None, None

    def token_first(self, skip_ws=True, skip_cm=False):
        """Returns the first child token.

        If *skip_ws* is ``True`` (the default), whitespace
        tokens are ignored.

        if *skip_cm* is ``True`` (default: ``False``), comments are
        ignored too.
        """

        # this on is inconsistent, using Comment instead of T.Comment...
        def matcher(tk):
            return not (
                (skip_ws and tk.is_whitespace)
                or (skip_cm and imt(tk, t=T.Comment, i=Comment))
            )

        return self._token_matching(matcher)[1]

    def token_next_by(self, i=None, m=None, t=None, idx=-1, end=None):
        idx += 1
        return self._token_matching(lambda tk: imt(tk, i, m, t), idx, end)

    def token_not_matching(self, funcs, idx):
        funcs = (funcs,) if not isinstance(funcs, (list, tuple)) else funcs
        funcs = [lambda tk: not func(tk) for func in funcs]
        return self._token_matching(funcs, idx)

    def token_matching(self, funcs, idx):
        return self._token_matching(funcs, idx)[1]

    def token_prev(self, idx, skip_ws=True, skip_cm=False):
        """Returns the previous token relative to *idx*.

        If *skip_ws* is ``True`` (the default) whitespace tokens are ignored.
        If *skip_cm* is ``True`` comments are ignored.
        ``None`` is returned if there's no previous token.
        """
        return self.token_next(idx, skip_ws, skip_cm, _reverse=True)

    # TODO: May need to re-add default value to idx
    def token_next(self, idx, skip_ws=True, skip_cm=False, _reverse=False):
        """Returns the next token relative to *idx*.

        If *skip_ws* is ``True`` (the default) whitespace tokens are ignored.
        If *skip_cm* is ``True`` comments are ignored.
        ``None`` is returned if there's no next token.
        """
        if idx is None:
            return None, None
        idx += 1  # alot of code usage current pre-compensates for this

        def matcher(tk):
            return not (
                (skip_ws and tk.is_whitespace)
                or (skip_cm and imt(tk, t=T.Comment, i=Comment))
            )

        return self._token_matching(matcher, idx, reverse=_reverse)

    def token_index(self, token, start=0):
        """Return list index of token."""
        start = start if isinstance(start, int) else self.token_index(start)
        return start + self.tokens[start:].index(token)

    def group_tokens(self, grp_cls, start, end, include_end=True, extend=False):
        """Replace tokens by an instance of *grp_cls*."""
        start_idx = start
        start = self.tokens[start_idx]

        end_idx = end + include_end

        # will be needed later for new group_clauses
        # while skip_ws and tokens and tokens[-1].is_whitespace:
        #     tokens = tokens[:-1]

        if extend and isinstance(start, grp_cls):
            subtokens = self.tokens[start_idx + 1 : end_idx]

            grp = start
            grp.tokens.extend(subtokens)
            del self.tokens[start_idx + 1 : end_idx]
            grp.value = str(start)
        else:
            subtokens = self.tokens[start_idx:end_idx]
            grp = grp_cls(subtokens)
            self.tokens[start_idx:end_idx] = [grp]
            grp.parent = self

        for token in subtokens:
            token.parent = grp

        return grp

    def insert_before(self, where, token):
        """Inserts *token* before *where*."""
        if not isinstance(where, int):
            where = self.token_index(where)
        token.parent = self
        self.tokens.insert(where, token)

    def insert_after(self, where, token, skip_ws=True):
        """Inserts *token* after *where*."""
        if not isinstance(where, int):
            where = self.token_index(where)
        nidx, next_ = self.token_next(where, skip_ws=skip_ws)
        token.parent = self
        if next_ is None:
            self.tokens.append(token)
        else:
            self.tokens.insert(nidx, token)

    def has_alias(self):
        """Returns ``True`` if an alias is present."""
        return self.get_alias() is not None

    def get_alias(self):
        """Returns the alias for this identifier or ``None``."""
        return None

    def get_name(self):
        """Returns the name of this identifier.

        This is either it's alias or it's real name. The returned valued can
        be considered as the name under which the object corresponding to
        this identifier is known within the current statement.
        """
        return self.get_alias() or self.get_real_name()

    def get_real_name(self):
        """Returns the real name (object name) of this identifier."""
        return None

    def get_parent_name(self):
        """Return name of the parent object if any.

        A parent object is identified by the first occurring dot.
        """
        dot_idx, _ = self.token_next_by(m=(T.Punctuation, "."))
        _, prev_ = self.token_prev(dot_idx)
        return remove_quotes(prev_.value) if prev_ is not None else None

    def _get_first_name(self, idx=None, reverse=False, keywords=False, real_name=False):
        """Returns the name of the first token with a name"""

        tokens = self.tokens[idx:] if idx else self.tokens
        tokens = reversed(tokens) if reverse else tokens
        types = [T.Name, T.Wildcard, T.String.Symbol]

        if keywords:
            types.append(T.Keyword)

        for token in tokens:
            if token.ttype in types:
                return remove_quotes(token.value)
            elif isinstance(token, (Identifier, Function)):
                return token.get_real_name() if real_name else token.get_name()


class Statement(TokenList):
    """Represents a SQL statement."""

    def get_type(self):
        """Returns the type of a statement.

        The returned value is a string holding an upper-cased reprint of
        the first DML or DDL keyword. If the first token in this group
        isn't a DML or DDL keyword "UNKNOWN" is returned.

        Whitespaces and comments at the beginning of the statement
        are ignored.
        """
        token = self.token_first(skip_cm=True)
        if token is None:
            # An "empty" statement that either has not tokens at all
            # or only whitespace tokens.
            return "UNKNOWN"

        elif token.ttype in (T.Keyword.DML, T.Keyword.DDL):
            return token.normalized

        elif token.ttype == T.Keyword.CTE:
            # The WITH keyword should be followed by either an Identifier or
            # an IdentifierList containing the CTE definitions;  the actual
            # DML keyword (e.g. SELECT, INSERT) will follow next.
            tidx = self.token_index(token)
            while tidx is not None:
                tidx, token = self.token_next(tidx, skip_ws=True)
                if isinstance(token, (Identifier, IdentifierList)):
                    tidx, token = self.token_next(tidx, skip_ws=True)

                    if token is not None and token.ttype == T.Keyword.DML:
                        return token.normalized

        # Hmm, probably invalid syntax, so return unknown.
        return "UNKNOWN"


class Identifier(NameAliasMixin, TokenList):
    """Represents an identifier.

    Identifiers may have aliases or typecasts.
    """

    def is_wildcard(self):
        """Return ``True`` if this identifier contains a wildcard."""
        _, token = self.token_next_by(t=T.Wildcard)
        return token is not None

    def get_typecast(self):
        """Returns the typecast or ``None`` of this object as a string."""
        midx, marker = self.token_next_by(m=(T.Punctuation, "::"))
        nidx, next_ = self.token_next(midx, skip_ws=False)
        return next_.value if next_ else None

    def get_ordering(self):
        """Returns the ordering or ``None`` as uppercase string."""
        _, ordering = self.token_next_by(t=T.Keyword.Order)
        return ordering.normalized if ordering else None

    def get_array_indices(self):
        """Returns an iterator of index token lists"""

        for token in self.tokens:
            if isinstance(token, SquareBrackets):
                # Use [1:-1] index to discard the square brackets
                yield token.tokens[1:-1]


class IdentifierList(TokenList):
    """A list of :class:`~sqlparse.sql.Identifier`\'s."""

    def get_identifiers(self):
        """Returns the identifiers.

        Whitespaces and punctuations are not included in this generator.
        """
        for token in self.tokens:
            if not (token.is_whitespace or token.match(T.Punctuation, ",")):
                yield token


class TypedLiteral(TokenList):
    """A typed literal, such as "date '2001-09-28'" or "interval '2 hours'"."""

    M_OPEN = [(T.Name.Builtin, None), (T.Keyword, "TIMESTAMP")]
    M_CLOSE = T.String.Single, None
    M_EXTEND = T.Keyword, ("DAY", "HOUR", "MINUTE", "MONTH", "SECOND", "YEAR")


class Parenthesis(TokenList):
    """Tokens between parenthesis."""

    M_OPEN = T.Punctuation, "("
    M_CLOSE = T.Punctuation, ")"

    @property
    def _groupable_tokens(self):
        return self.tokens[1:-1]


class SquareBrackets(TokenList):
    """Tokens between square brackets"""

    M_OPEN = T.Punctuation, "["
    M_CLOSE = T.Punctuation, "]"

    @property
    def _groupable_tokens(self):
        return self.tokens[1:-1]


class Assignment(TokenList):
    """An assignment like 'var := val;'"""


class If(TokenList):
    """An 'if' clause with possible 'else if' or 'else' parts."""

    M_OPEN = T.Keyword, "IF"
    M_CLOSE = T.Keyword, "END IF"


class For(TokenList):
    """A 'FOR' loop."""

    M_OPEN = T.Keyword, ("FOR", "FOREACH")
    M_CLOSE = T.Keyword, "END LOOP"


class Comparison(TokenList):
    """A comparison used for example in WHERE clauses."""

    @property
    def left(self):
        return self.tokens[0]

    @property
    def right(self):
        return self.tokens[-1]


class Comment(TokenList):
    """A comment."""

    def is_multiline(self):
        return self.tokens and self.tokens[0].ttype == T.Comment.Multiline


class Where(TokenList):
    """A WHERE clause."""

    M_OPEN = T.Keyword, "WHERE"
    M_CLOSE = T.Keyword, (
        "ORDER BY",
        "GROUP BY",
        "LIMIT",
        "UNION",
        "UNION ALL",
        "EXCEPT",
        "HAVING",
        "RETURNING",
        "INTO",
    )


class Having(TokenList):
    """A HAVING clause."""

    M_OPEN = T.Keyword, "HAVING"
    M_CLOSE = T.Keyword, ("ORDER BY", "LIMIT")


class Case(TokenList):
    """A CASE statement with one or more WHEN and possibly an ELSE part."""

    M_OPEN = T.Keyword, "CASE"
    M_CLOSE = T.Keyword, "END"

    def get_cases(self, skip_ws=False):
        """Returns a list of 2-tuples (condition, value).

        If an ELSE exists condition is None.
        """
        CONDITION = 1
        VALUE = 2

        ret = []
        mode = CONDITION

        for token in self.tokens:
            # Set mode from the current statement
            if token.match(T.Keyword, "CASE"):
                continue

            elif skip_ws and token.ttype in T.Whitespace:
                continue

            elif token.match(T.Keyword, "WHEN"):
                ret.append(([], []))
                mode = CONDITION

            elif token.match(T.Keyword, "THEN"):
                mode = VALUE

            elif token.match(T.Keyword, "ELSE"):
                ret.append((None, []))
                mode = VALUE

            elif token.match(T.Keyword, "END"):
                mode = None

            # First condition without preceding WHEN
            if mode and not ret:
                ret.append(([], []))

            # Append token depending of the current mode
            if mode == CONDITION:
                ret[-1][0].append(token)

            elif mode == VALUE:
                ret[-1][1].append(token)

        # Return cases list
        return ret


class Function(NameAliasMixin, TokenList):
    """A function or procedure call."""

    def get_parameters(self):
        """Return a list of parameters."""
        parenthesis = self.tokens[-1]
        for token in parenthesis.tokens:
            if isinstance(token, IdentifierList):
                return token.get_identifiers()
            elif imt(token, i=(Function, Identifier), t=T.Literal):
                return [
                    token,
                ]
        return []


class Begin(TokenList):
    """A BEGIN/END block."""

    M_OPEN = T.Keyword, "BEGIN"
    M_CLOSE = T.Keyword, "END"


class Operation(TokenList):
    """Grouping of operations"""


class Values(TokenList):
    """Grouping of values"""


class Command(TokenList):
    """Grouping of CLI commands."""



import unittest

from spetlr import Configurator
from tests.cluster.sql.DeliverySqlExecutor import DeliverySqlExecutor
from tests.cluster.sql.DeliverySqlServer import DeliverySqlServer

from . import extras


class DeliverySqlExecutorTests(unittest.TestCase):
    tc = None
    sql_server = None

    @classmethod
    def setUpClass(cls):
        cls.sql_server = DeliverySqlServer()

        # Register the delivery table for the table configurator
        cls.tc = Configurator()
        cls.tc.add_resource_path(extras)
        cls.tc.set_debug()

        # Ensure no table is there
        cls.sql_server.drop_table("SqlTestTable1")
        cls.sql_server.drop_table("SqlTestTable2")

    @classmethod
    def tearDownClass(cls):
        cls.sql_server.drop_table("SqlTestTable1")
        cls.sql_server.drop_table("SqlTestTable2")

    def test_can_execute(self):
        DeliverySqlExecutor().execute_sql_file("test1")

        self.sql_server.read_table("SqlTestTable1")

        sh = self.sql_server.from_tc("SqlTestTable2")
        sh.read()

        self.assertTrue(True)



import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.configurator import Configurator
from spetlr.delta import DeltaHandle
from spetlr.schema_manager import SchemaManager
from spetlr.schema_manager.spark_schema import get_schema

from . import extras
from .SparkExecutor import SparkSqlExecutor


class TestSchemaManager(DataframeTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        c = Configurator()
        c.clear_all_configurations()
        c.add_resource_path(extras)
        SchemaManager().clear_all_configurations()
        extras.initSchemaManager()

    def test_get_sql_schema(self):
        schema = SchemaManager().get_schema(schema_identifier="SchemaTestTable2")

        expected_schema = T.StructType(
            [
                T.StructField("a", T.IntegerType(), True),
                T.StructField("b", T.StringType(), True),
            ]
        )

        self.assertEqualSchema(schema, expected_schema)

    def test_get_schema_as_string(self):
        schema = SchemaManager().get_schema_as_string(
            schema_identifier="SchemaTestTable2"
        )

        self.assertEqual(schema, "a int, b string")

    def test_get_all_schemas(self):
        schemas_dict = SchemaManager().get_all_schemas()

        expected_schemas = {
            "python_test_schema": extras.python_test_schema,
            "python_test_schema2": extras.python_test_schema2,
            "SchemaTestTable1": extras.python_test_schema2,
            "SchemaTestTable2": T.StructType(
                [
                    T.StructField("a", T.IntegerType(), True),
                    T.StructField("b", T.StringType(), True),
                ]
            ),
        }

        self.assertDictEqual(expected_schemas, schemas_dict)

    def test_get_all_spark_sql_schemas(self):
        schemas_dict = SchemaManager().get_all_spark_sql_schemas()

        test_schema1_str = (
            "a int, "
            'b int  COMMENT "really? is that it?", '
            "c string, "
            "cplx struct<someId:string,details:struct<id:string>,blabla:array<int>>, "
            "d timestamp, "
            "m map<int,string>, "
            "p decimal(10,3), "
            "final string"
        )
        test_schema2_str = (
            "a int, "
            "c string, "
            "d timestamp, "
            "m map<int,string>, "
            "p decimal(10,3), "
            "final string"
        )
        expected_schemas = {
            "python_test_schema": test_schema1_str,
            "python_test_schema2": test_schema2_str,
            "SchemaTestTable1": test_schema2_str,
            "SchemaTestTable2": "a int, b string",
        }
        self.maxDiff = None
        self.assertDictEqual(expected_schemas, schemas_dict)

    def test_schema_to_spark_sql(self):
        schema = T.StructType(
            [
                T.StructField("Column1", T.IntegerType(), True),
                T.StructField("Column2", T.StringType(), True),
                T.StructField("Column3", T.FloatType(), True),
            ]
        )

        transformed_str = SchemaManager()._schema_to_spark_sql(schema=schema)
        expected_str = "Column1 int, Column2 string, Column3 float"

        self.maxDiff = None
        self.assertEqual(expected_str, transformed_str)

        transformed_schema = get_schema(transformed_str)
        self.maxDiff = None
        self.assertEqualSchema(schema, transformed_schema)

    def test_sql_executor_schema(self):
        SparkSqlExecutor().execute_sql_file("*")

        test_df = DeltaHandle.from_tc("SchemaTestTable1").read()

        self.maxDiff = None
        self.assertEqualSchema(
            SchemaManager().get_schema("python_test_schema2"), test_df.schema
        )



import itertools
import re
from importlib import resources as ir
from pathlib import Path
from types import ModuleType
from typing import Dict, List, Optional, Union

from spetlr.configurator.configurator import Configurator
from spetlr.configurator.sql import sqlparse
from spetlr.configurator.sql.init_sqlparse import parse
from spetlr.schema_manager import SchemaManager
from spetlr.spark import Spark
from spetlr.sql import BaseExecutor


class SqlExecutor:
    _DEFAULT = object()

    def __init__(
        self,
        base_module: Union[str, ModuleType] = None,
        server: BaseExecutor = None,
        statement_spliter: Optional[List[str]] = _DEFAULT,
    ):
        """Class to pre-treat sql statements and execute them.
        Replacement sequenced related to the Configurator will be inserted before
        execution.
        Giving not statement_spliter is equivalent to [";", "-- COMMAND ----------"].
        To disable splitting statements, please supply statement_spliter = None.
        Semicolon will be treated correctly when quoted or in comments.

        Default behavior supports spark, which will complain if given
        no actual sql code or on multiple statements."""

        self.base_module = base_module
        self.server = server
        if statement_spliter is self._DEFAULT:
            # the sequence "-- COMMAND ----------" is used in jupyter notebooks
            # and separates cells.
            # We treat it as another way to end a statement
            statement_spliter = [";", "-- COMMAND ----------"]
        self.statement_spliter = statement_spliter

    def _wildcard_string_to_regexp(self, instr: str) -> str:
        # prepare file pattern:
        if instr.endswith(".sql"):
            instr = instr[:-4]

        instr = instr.replace("*", ".*")

        # the string end indicator will prevent us from matching the wrong files
        #  where a filename is also a prefix to another filename
        if not instr.endswith("$"):
            instr = instr + "$"

        return instr

    def get_statements(
        self,
        file_pattern: str,
        exclude_pattern: str = None,
        replacements: Dict[str, str] = None,
    ):
        """
        NB: This sql parser can be challenged in parsing sql statements
        which do not use semicolon as a query separator only.
        """

        # prepare arguments:
        file_pattern = self._wildcard_string_to_regexp(file_pattern)
        if exclude_pattern is not None:
            exclude_pattern = self._wildcard_string_to_regexp(exclude_pattern)
        if replacements is None:
            replacements = {}

        # prepare the full set of replacements
        schema_replacements = {
            f"{k}_schema": v
            for k, v in SchemaManager().get_all_spark_sql_schemas().items()
        }
        replacements = {
            **(Configurator().get_all_details()),
            **schema_replacements,
            **replacements,
        }

        # loop the module contents and find matching files
        for file_name in ir.contents(self.base_module):
            extension = Path(file_name).suffix
            if extension not in [".sql"]:
                continue

            if not re.match(file_pattern, Path(file_name).stem):
                continue

            if exclude_pattern is not None and re.search(
                exclude_pattern, Path(file_name).stem
            ):
                continue

            with ir.path(self.base_module, file_name) as file_path:
                with open(file_path) as file:
                    conts = file.read()
                    sql_code = conts.format(**replacements)

                    if self.statement_spliter is None:
                        code_parts = [sql_code]
                    elif ";" not in self.statement_spliter:
                        code_parts = [sql_code]
                        for sequence in self.statement_spliter:
                            code_parts = itertools.chain.from_iterable(
                                part.split(sequence) for part in code_parts
                            )

                    else:
                        # if ; is included split the file into sql statements
                        # by using parse, we ensure not to split by escaped or commented
                        # occurrences of ;

                        for marker in self.statement_spliter:
                            if marker != ";":
                                sql_code = sql_code.replace(marker, ";")

                        code_parts = [
                            "".join(token.value for token in statement)
                            for statement in parse(sql_code)
                        ]

                    for full_statement in code_parts:
                        cleaned_statement = (
                            (
                                "".join(
                                    token.value
                                    for statement in parse(full_statement)
                                    for token in statement
                                    if token.ttype not in sqlparse.tokens.Comment
                                )
                            )
                            .strip()
                            .strip(";")
                        )

                        # skip the statement unless it actually contains code.
                        # spark.sql complains if you only give it comments
                        if cleaned_statement:
                            yield full_statement

    def execute_sql_file(self, file_pattern: str, exclude_pattern: str = None):
        """
        NB: This sql parser can be challenged in parsing sql statements
        which do not use semicolon as a query separator only.
        """

        executor = self.server or Spark.get()

        for statement in self.get_statements(file_pattern, exclude_pattern):
            executor.sql(statement)



from functools import reduce

from pyspark.sql import DataFrame

from spetlr.etl import Transformer
from spetlr.etl.types import dataset_group


class UnionTransformer(Transformer):
    """
    This transformer unions multiple DataFrames

    Attributes:
    ----------
        allowMissingColumns : bool
            When the parameter allowMissingColumns is True,
            the set of column names in this and other DataFrame
            can differ; missing columns will be filled with null

    Methods
    -------
    process(df: DataFrame):
        returns the input DataFrame

    process_many(datasets: dataset_group):
        returns the union of all input DataFrames
    """

    def __init__(self, allowMissingColumns: bool = False):
        super().__init__()
        self.allowMissingColumns = allowMissingColumns

    def process(self, df: DataFrame) -> DataFrame:
        return df

    def process_many(self, datasets: dataset_group) -> DataFrame:
        dfs = list(datasets.values())
        return reduce(
            lambda df1, df2: df1.unionByName(
                df2, allowMissingColumns=self.allowMissingColumns
            ),
            dfs,
        )



#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

from spetlr.configurator.sql.sqlparse import tokens

# object() only supports "is" and is useful as a marker
# use this marker to specify that the given regex in SQL_REGEX
# shall be processed further through a lookup in the KEYWORDS dictionaries
PROCESS_AS_KEYWORD = object()


SQL_REGEX = [
    (r"(--|# )\+.*?(\r\n|\r|\n|$)", tokens.Comment.Single.Hint),
    (r"/\*\+[\s\S]*?\*/", tokens.Comment.Multiline.Hint),
    (r"(--|# ).*?(\r\n|\r|\n|$)", tokens.Comment.Single),
    (r"/\*[\s\S]*?\*/", tokens.Comment.Multiline),
    (r"(\r\n|\r|\n)", tokens.Newline),
    (r"\s+?", tokens.Whitespace),
    (r":=", tokens.Assignment),
    (r"::", tokens.Punctuation),
    (r"\*", tokens.Wildcard),
    (r"`(``|[^`])*`", tokens.Name),
    (r"´(´´|[^´])*´", tokens.Name),
    (r"((?<!\S)\$(?:[_A-ZÀ-Ü]\w*)?\$)[\s\S]*?\1", tokens.Literal),
    (r"\?", tokens.Name.Placeholder),
    (r"%(\(\w+\))?s", tokens.Name.Placeholder),
    (r"(?<!\w)[$:?]\w+", tokens.Name.Placeholder),
    (r"\\\w+", tokens.Command),
    # FIXME(andi): VALUES shouldn't be listed here
    # see https://github.com/andialbrecht/sqlparse/pull/64
    # AS and IN are special, it may be followed by a parenthesis, but
    # are never functions, see issue183 and issue507
    (r"(CASE|IN|VALUES|USING|FROM|AS)\b", tokens.Keyword),
    (r"(@|##|#)[A-ZÀ-Ü]\w+", tokens.Name),
    # see issue #39
    # Spaces around period `schema . name` are valid identifier
    # TODO: Spaces before period not implemented
    (r"[A-ZÀ-Ü]\w*(?=\s*\.)", tokens.Name),  # 'Name'.
    # FIXME(atronah): never match,
    # because `re.match` doesn't work with look-behind regexp feature
    (r"(?<=\.)[A-ZÀ-Ü]\w*", tokens.Name),  # .'Name'
    (r"[A-ZÀ-Ü]\w*(?=\()", tokens.Name),  # side effect: change kw to func
    (r"-?0x[\dA-F]+", tokens.Number.Hexadecimal),
    (r"-?\d+(\.\d+)?E-?\d+", tokens.Number.Float),
    (r"(?![_A-ZÀ-Ü])-?(\d+(\.\d*)|\.\d+)(?![_A-ZÀ-Ü])", tokens.Number.Float),
    (r"(?![_A-ZÀ-Ü])-?\d+(?![_A-ZÀ-Ü])", tokens.Number.Integer),
    (r"'(''|\\\\|\\'|[^'])*'", tokens.String.Single),
    # not a real string literal in ANSI SQL:
    (r'"(""|\\\\|\\"|[^"])*"', tokens.String.Symbol),
    (r'(""|".*?[^\\]")', tokens.String.Symbol),
    # sqlite names can be escaped with [square brackets]. left bracket
    # cannot be preceded by word character or a right bracket --
    # otherwise it's probably an array index
    (r"(?<![\w\])])(\[[^\]\[]+\])", tokens.Name),
    (
        r"((LEFT\s+|RIGHT\s+|FULL\s+)?(INNER\s+|OUTER\s+|STRAIGHT\s+)?"
        r"|(CROSS\s+|NATURAL\s+)?)?JOIN\b",
        tokens.Keyword,
    ),
    (r"END(\s+IF|\s+LOOP|\s+WHILE)?\b", tokens.Keyword),
    (r"NOT\s+NULL\b", tokens.Keyword),
    (r"NULLS\s+(FIRST|LAST)\b", tokens.Keyword),
    (r"UNION\s+ALL\b", tokens.Keyword),
    (r"CREATE(\s+OR\s+REPLACE)?\b", tokens.Keyword.DDL),
    (r"DOUBLE\s+PRECISION\b", tokens.Name.Builtin),
    (r"GROUP\s+BY\b", tokens.Keyword),
    (r"ORDER\s+BY\b", tokens.Keyword),
    (r"HANDLER\s+FOR\b", tokens.Keyword),
    (
        r"(LATERAL\s+VIEW\s+)" r"(EXPLODE|INLINE|PARSE_URL_TUPLE|POSEXPLODE|STACK)\b",
        tokens.Keyword,
    ),
    (r"(AT|WITH')\s+TIME\s+ZONE\s+'[^']+'", tokens.Keyword.TZCast),
    (r"(NOT\s+)?(LIKE|ILIKE|RLIKE)\b", tokens.Operator.Comparison),
    (r"(NOT\s+)?(REGEXP)\b", tokens.Operator.Comparison),
    # Check for keywords, also returns tokens.Name if regex matches
    # but the match isn't a keyword.
    (r"[0-9_\w][_$#\w]*", PROCESS_AS_KEYWORD),
    (r"[;:()\[\],\.]", tokens.Punctuation),
    (r"[<>=~!]+", tokens.Operator.Comparison),
    (r"[+/@#%^&|^-]+", tokens.Operator),
]

KEYWORDS = {
    "ABORT": tokens.Keyword,
    "ABS": tokens.Keyword,
    "ABSOLUTE": tokens.Keyword,
    "ACCESS": tokens.Keyword,
    "ADA": tokens.Keyword,
    "ADD": tokens.Keyword,
    "ADMIN": tokens.Keyword,
    "AFTER": tokens.Keyword,
    "AGGREGATE": tokens.Keyword,
    "ALIAS": tokens.Keyword,
    "ALL": tokens.Keyword,
    "ALLOCATE": tokens.Keyword,
    "ANALYSE": tokens.Keyword,
    "ANALYZE": tokens.Keyword,
    "ANY": tokens.Keyword,
    "ARRAYLEN": tokens.Keyword,
    "ARE": tokens.Keyword,
    "ASC": tokens.Keyword.Order,
    "ASENSITIVE": tokens.Keyword,
    "ASSERTION": tokens.Keyword,
    "ASSIGNMENT": tokens.Keyword,
    "ASYMMETRIC": tokens.Keyword,
    "AT": tokens.Keyword,
    "ATOMIC": tokens.Keyword,
    "AUDIT": tokens.Keyword,
    "AUTHORIZATION": tokens.Keyword,
    "AUTO_INCREMENT": tokens.Keyword,
    "AVG": tokens.Keyword,
    "BACKWARD": tokens.Keyword,
    "BEFORE": tokens.Keyword,
    "BEGIN": tokens.Keyword,
    "BETWEEN": tokens.Keyword,
    "BITVAR": tokens.Keyword,
    "BIT_LENGTH": tokens.Keyword,
    "BOTH": tokens.Keyword,
    "BREADTH": tokens.Keyword,
    # 'C': tokens.Keyword,  # most likely this is an alias
    "CACHE": tokens.Keyword,
    "CALL": tokens.Keyword,
    "CALLED": tokens.Keyword,
    "CARDINALITY": tokens.Keyword,
    "CASCADE": tokens.Keyword,
    "CASCADED": tokens.Keyword,
    "CAST": tokens.Keyword,
    "CATALOG": tokens.Keyword,
    "CATALOG_NAME": tokens.Keyword,
    "CHAIN": tokens.Keyword,
    "CHARACTERISTICS": tokens.Keyword,
    "CHARACTER_LENGTH": tokens.Keyword,
    "CHARACTER_SET_CATALOG": tokens.Keyword,
    "CHARACTER_SET_NAME": tokens.Keyword,
    "CHARACTER_SET_SCHEMA": tokens.Keyword,
    "CHAR_LENGTH": tokens.Keyword,
    "CHARSET": tokens.Keyword,
    "CHECK": tokens.Keyword,
    "CHECKED": tokens.Keyword,
    "CHECKPOINT": tokens.Keyword,
    "CLASS": tokens.Keyword,
    "CLASS_ORIGIN": tokens.Keyword,
    "CLOB": tokens.Keyword,
    "CLOSE": tokens.Keyword,
    "CLUSTER": tokens.Keyword,
    "COALESCE": tokens.Keyword,
    "COBOL": tokens.Keyword,
    "COLLATE": tokens.Keyword,
    "COLLATION": tokens.Keyword,
    "COLLATION_CATALOG": tokens.Keyword,
    "COLLATION_NAME": tokens.Keyword,
    "COLLATION_SCHEMA": tokens.Keyword,
    "COLLECT": tokens.Keyword,
    "COLUMN": tokens.Keyword,
    "COLUMN_NAME": tokens.Keyword,
    "COMPRESS": tokens.Keyword,
    "COMMAND_FUNCTION": tokens.Keyword,
    "COMMAND_FUNCTION_CODE": tokens.Keyword,
    "COMMENT": tokens.Keyword,
    "COMMIT": tokens.Keyword.DML,
    "COMMITTED": tokens.Keyword,
    "COMPLETION": tokens.Keyword,
    "CONCURRENTLY": tokens.Keyword,
    "CONDITION_NUMBER": tokens.Keyword,
    "CONNECT": tokens.Keyword,
    "CONNECTION": tokens.Keyword,
    "CONNECTION_NAME": tokens.Keyword,
    "CONSTRAINT": tokens.Keyword,
    "CONSTRAINTS": tokens.Keyword,
    "CONSTRAINT_CATALOG": tokens.Keyword,
    "CONSTRAINT_NAME": tokens.Keyword,
    "CONSTRAINT_SCHEMA": tokens.Keyword,
    "CONSTRUCTOR": tokens.Keyword,
    "CONTAINS": tokens.Keyword,
    "CONTINUE": tokens.Keyword,
    "CONVERSION": tokens.Keyword,
    "CONVERT": tokens.Keyword,
    "COPY": tokens.Keyword,
    "CORRESPONDING": tokens.Keyword,
    "COUNT": tokens.Keyword,
    "CREATEDB": tokens.Keyword,
    "CREATEUSER": tokens.Keyword,
    "CROSS": tokens.Keyword,
    "CUBE": tokens.Keyword,
    "CURRENT": tokens.Keyword,
    "CURRENT_DATE": tokens.Keyword,
    "CURRENT_PATH": tokens.Keyword,
    "CURRENT_ROLE": tokens.Keyword,
    "CURRENT_TIME": tokens.Keyword,
    "CURRENT_TIMESTAMP": tokens.Keyword,
    "CURRENT_USER": tokens.Keyword,
    "CURSOR": tokens.Keyword,
    "CURSOR_NAME": tokens.Keyword,
    "CYCLE": tokens.Keyword,
    "DATA": tokens.Keyword,
    "DATABASE": tokens.Keyword,
    "DATETIME_INTERVAL_CODE": tokens.Keyword,
    "DATETIME_INTERVAL_PRECISION": tokens.Keyword,
    "DAY": tokens.Keyword,
    "DEALLOCATE": tokens.Keyword,
    "DECLARE": tokens.Keyword,
    "DEFAULT": tokens.Keyword,
    "DEFAULTS": tokens.Keyword,
    "DEFERRABLE": tokens.Keyword,
    "DEFERRED": tokens.Keyword,
    "DEFINED": tokens.Keyword,
    "DEFINER": tokens.Keyword,
    "DELIMITER": tokens.Keyword,
    "DELIMITERS": tokens.Keyword,
    "DEREF": tokens.Keyword,
    "DESC": tokens.Keyword.Order,
    "DESCRIBE": tokens.Keyword,
    "DESCRIPTOR": tokens.Keyword,
    "DESTROY": tokens.Keyword,
    "DESTRUCTOR": tokens.Keyword,
    "DETERMINISTIC": tokens.Keyword,
    "DIAGNOSTICS": tokens.Keyword,
    "DICTIONARY": tokens.Keyword,
    "DISABLE": tokens.Keyword,
    "DISCONNECT": tokens.Keyword,
    "DISPATCH": tokens.Keyword,
    "DIV": tokens.Operator,
    "DO": tokens.Keyword,
    "DOMAIN": tokens.Keyword,
    "DYNAMIC": tokens.Keyword,
    "DYNAMIC_FUNCTION": tokens.Keyword,
    "DYNAMIC_FUNCTION_CODE": tokens.Keyword,
    "EACH": tokens.Keyword,
    "ENABLE": tokens.Keyword,
    "ENCODING": tokens.Keyword,
    "ENCRYPTED": tokens.Keyword,
    "END-EXEC": tokens.Keyword,
    "ENGINE": tokens.Keyword,
    "EQUALS": tokens.Keyword,
    "ESCAPE": tokens.Keyword,
    "EVERY": tokens.Keyword,
    "EXCEPT": tokens.Keyword,
    "EXCEPTION": tokens.Keyword,
    "EXCLUDING": tokens.Keyword,
    "EXCLUSIVE": tokens.Keyword,
    "EXEC": tokens.Keyword,
    "EXECUTE": tokens.Keyword,
    "EXISTING": tokens.Keyword,
    "EXISTS": tokens.Keyword,
    "EXPLAIN": tokens.Keyword,
    "EXTERNAL": tokens.Keyword,
    "EXTRACT": tokens.Keyword,
    "FALSE": tokens.Keyword,
    "FETCH": tokens.Keyword,
    "FILE": tokens.Keyword,
    "FINAL": tokens.Keyword,
    "FIRST": tokens.Keyword,
    "FORCE": tokens.Keyword,
    "FOREACH": tokens.Keyword,
    "FOREIGN": tokens.Keyword,
    "FORTRAN": tokens.Keyword,
    "FORWARD": tokens.Keyword,
    "FOUND": tokens.Keyword,
    "FREE": tokens.Keyword,
    "FREEZE": tokens.Keyword,
    "FULL": tokens.Keyword,
    "FUNCTION": tokens.Keyword,
    # 'G': tokens.Keyword,
    "GENERAL": tokens.Keyword,
    "GENERATED": tokens.Keyword,
    "GET": tokens.Keyword,
    "GLOBAL": tokens.Keyword,
    "GO": tokens.Keyword,
    "GOTO": tokens.Keyword,
    "GRANT": tokens.Keyword,
    "GRANTED": tokens.Keyword,
    "GROUPING": tokens.Keyword,
    "HAVING": tokens.Keyword,
    "HIERARCHY": tokens.Keyword,
    "HOLD": tokens.Keyword,
    "HOUR": tokens.Keyword,
    "HOST": tokens.Keyword,
    "IDENTIFIED": tokens.Keyword,
    "IDENTITY": tokens.Keyword,
    "IGNORE": tokens.Keyword,
    "ILIKE": tokens.Keyword,
    "IMMEDIATE": tokens.Keyword,
    "IMMUTABLE": tokens.Keyword,
    "IMPLEMENTATION": tokens.Keyword,
    "IMPLICIT": tokens.Keyword,
    "INCLUDING": tokens.Keyword,
    "INCREMENT": tokens.Keyword,
    "INDEX": tokens.Keyword,
    "INDICATOR": tokens.Keyword,
    "INFIX": tokens.Keyword,
    "INHERITS": tokens.Keyword,
    "INITIAL": tokens.Keyword,
    "INITIALIZE": tokens.Keyword,
    "INITIALLY": tokens.Keyword,
    "INOUT": tokens.Keyword,
    "INPUT": tokens.Keyword,
    "INSENSITIVE": tokens.Keyword,
    "INSTANTIABLE": tokens.Keyword,
    "INSTEAD": tokens.Keyword,
    "INTERSECT": tokens.Keyword,
    "INTO": tokens.Keyword,
    "INVOKER": tokens.Keyword,
    "IS": tokens.Keyword,
    "ISNULL": tokens.Keyword,
    "ISOLATION": tokens.Keyword,
    "ITERATE": tokens.Keyword,
    # 'K': tokens.Keyword,
    "KEY": tokens.Keyword,
    "KEY_MEMBER": tokens.Keyword,
    "KEY_TYPE": tokens.Keyword,
    "LANCOMPILER": tokens.Keyword,
    "LANGUAGE": tokens.Keyword,
    "LARGE": tokens.Keyword,
    "LAST": tokens.Keyword,
    "LATERAL": tokens.Keyword,
    "LEADING": tokens.Keyword,
    "LENGTH": tokens.Keyword,
    "LESS": tokens.Keyword,
    "LEVEL": tokens.Keyword,
    "LIMIT": tokens.Keyword,
    "LISTEN": tokens.Keyword,
    "LOAD": tokens.Keyword,
    "LOCAL": tokens.Keyword,
    "LOCALTIME": tokens.Keyword,
    "LOCALTIMESTAMP": tokens.Keyword,
    "LOCATION": tokens.Keyword,
    "LOCATOR": tokens.Keyword,
    "LOCK": tokens.Keyword,
    "LOWER": tokens.Keyword,
    # 'M': tokens.Keyword,
    "MAP": tokens.Keyword,
    "MATCH": tokens.Keyword,
    "MAXEXTENTS": tokens.Keyword,
    "MAXVALUE": tokens.Keyword,
    "MESSAGE_LENGTH": tokens.Keyword,
    "MESSAGE_OCTET_LENGTH": tokens.Keyword,
    "MESSAGE_TEXT": tokens.Keyword,
    "METHOD": tokens.Keyword,
    "MINUTE": tokens.Keyword,
    "MINUS": tokens.Keyword,
    "MINVALUE": tokens.Keyword,
    "MOD": tokens.Keyword,
    "MODE": tokens.Keyword,
    "MODIFIES": tokens.Keyword,
    "MODIFY": tokens.Keyword,
    "MONTH": tokens.Keyword,
    "MORE": tokens.Keyword,
    "MOVE": tokens.Keyword,
    "MUMPS": tokens.Keyword,
    "NAMES": tokens.Keyword,
    "NATIONAL": tokens.Keyword,
    "NATURAL": tokens.Keyword,
    "NCHAR": tokens.Keyword,
    "NCLOB": tokens.Keyword,
    "NEW": tokens.Keyword,
    "NEXT": tokens.Keyword,
    "NO": tokens.Keyword,
    "NOAUDIT": tokens.Keyword,
    "NOCOMPRESS": tokens.Keyword,
    "NOCREATEDB": tokens.Keyword,
    "NOCREATEUSER": tokens.Keyword,
    "NONE": tokens.Keyword,
    "NOT": tokens.Keyword,
    "NOTFOUND": tokens.Keyword,
    "NOTHING": tokens.Keyword,
    "NOTIFY": tokens.Keyword,
    "NOTNULL": tokens.Keyword,
    "NOWAIT": tokens.Keyword,
    "NULL": tokens.Keyword,
    "NULLABLE": tokens.Keyword,
    "NULLIF": tokens.Keyword,
    "OBJECT": tokens.Keyword,
    "OCTET_LENGTH": tokens.Keyword,
    "OF": tokens.Keyword,
    "OFF": tokens.Keyword,
    "OFFLINE": tokens.Keyword,
    "OFFSET": tokens.Keyword,
    "OIDS": tokens.Keyword,
    "OLD": tokens.Keyword,
    "ONLINE": tokens.Keyword,
    "ONLY": tokens.Keyword,
    "OPEN": tokens.Keyword,
    "OPERATION": tokens.Keyword,
    "OPERATOR": tokens.Keyword,
    "OPTION": tokens.Keyword,
    "OPTIONS": tokens.Keyword,
    "ORDINALITY": tokens.Keyword,
    "OUT": tokens.Keyword,
    "OUTPUT": tokens.Keyword,
    "OVERLAPS": tokens.Keyword,
    "OVERLAY": tokens.Keyword,
    "OVERRIDING": tokens.Keyword,
    "OWNER": tokens.Keyword,
    "QUARTER": tokens.Keyword,
    "PAD": tokens.Keyword,
    "PARAMETER": tokens.Keyword,
    "PARAMETERS": tokens.Keyword,
    "PARAMETER_MODE": tokens.Keyword,
    "PARAMETER_NAME": tokens.Keyword,
    "PARAMETER_ORDINAL_POSITION": tokens.Keyword,
    "PARAMETER_SPECIFIC_CATALOG": tokens.Keyword,
    "PARAMETER_SPECIFIC_NAME": tokens.Keyword,
    "PARAMETER_SPECIFIC_SCHEMA": tokens.Keyword,
    "PARTIAL": tokens.Keyword,
    "PASCAL": tokens.Keyword,
    "PCTFREE": tokens.Keyword,
    "PENDANT": tokens.Keyword,
    "PLACING": tokens.Keyword,
    "PLI": tokens.Keyword,
    "POSITION": tokens.Keyword,
    "POSTFIX": tokens.Keyword,
    "PRECISION": tokens.Keyword,
    "PREFIX": tokens.Keyword,
    "PREORDER": tokens.Keyword,
    "PREPARE": tokens.Keyword,
    "PRESERVE": tokens.Keyword,
    "PRIMARY": tokens.Keyword,
    "PRIOR": tokens.Keyword,
    "PRIVILEGES": tokens.Keyword,
    "PROCEDURAL": tokens.Keyword,
    "PROCEDURE": tokens.Keyword,
    "PUBLIC": tokens.Keyword,
    "RAISE": tokens.Keyword,
    "RAW": tokens.Keyword,
    "READ": tokens.Keyword,
    "READS": tokens.Keyword,
    "RECHECK": tokens.Keyword,
    "RECURSIVE": tokens.Keyword,
    "REF": tokens.Keyword,
    "REFERENCES": tokens.Keyword,
    "REFERENCING": tokens.Keyword,
    "REINDEX": tokens.Keyword,
    "RELATIVE": tokens.Keyword,
    "RENAME": tokens.Keyword,
    "REPEATABLE": tokens.Keyword,
    "RESET": tokens.Keyword,
    "RESOURCE": tokens.Keyword,
    "RESTART": tokens.Keyword,
    "RESTRICT": tokens.Keyword,
    "RESULT": tokens.Keyword,
    "RETURN": tokens.Keyword,
    "RETURNED_LENGTH": tokens.Keyword,
    "RETURNED_OCTET_LENGTH": tokens.Keyword,
    "RETURNED_SQLSTATE": tokens.Keyword,
    "RETURNING": tokens.Keyword,
    "RETURNS": tokens.Keyword,
    "REVOKE": tokens.Keyword,
    "RIGHT": tokens.Keyword,
    "ROLE": tokens.Keyword,
    "ROLLBACK": tokens.Keyword.DML,
    "ROLLUP": tokens.Keyword,
    "ROUTINE": tokens.Keyword,
    "ROUTINE_CATALOG": tokens.Keyword,
    "ROUTINE_NAME": tokens.Keyword,
    "ROUTINE_SCHEMA": tokens.Keyword,
    "ROW": tokens.Keyword,
    "ROWS": tokens.Keyword,
    "ROW_COUNT": tokens.Keyword,
    "RULE": tokens.Keyword,
    "SAVE_POINT": tokens.Keyword,
    "SCALE": tokens.Keyword,
    "SCHEMA": tokens.Keyword,
    "SCHEMA_NAME": tokens.Keyword,
    "SCOPE": tokens.Keyword,
    "SCROLL": tokens.Keyword,
    "SEARCH": tokens.Keyword,
    "SECOND": tokens.Keyword,
    "SECURITY": tokens.Keyword,
    "SELF": tokens.Keyword,
    "SENSITIVE": tokens.Keyword,
    "SEQUENCE": tokens.Keyword,
    "SERIALIZABLE": tokens.Keyword,
    "SERVER_NAME": tokens.Keyword,
    "SESSION": tokens.Keyword,
    "SESSION_USER": tokens.Keyword,
    "SETOF": tokens.Keyword,
    "SETS": tokens.Keyword,
    "SHARE": tokens.Keyword,
    "SHOW": tokens.Keyword,
    "SIMILAR": tokens.Keyword,
    "SIMPLE": tokens.Keyword,
    "SIZE": tokens.Keyword,
    "SOME": tokens.Keyword,
    "SOURCE": tokens.Keyword,
    "SPACE": tokens.Keyword,
    "SPECIFIC": tokens.Keyword,
    "SPECIFICTYPE": tokens.Keyword,
    "SPECIFIC_NAME": tokens.Keyword,
    "SQL": tokens.Keyword,
    "SQLBUF": tokens.Keyword,
    "SQLCODE": tokens.Keyword,
    "SQLERROR": tokens.Keyword,
    "SQLEXCEPTION": tokens.Keyword,
    "SQLSTATE": tokens.Keyword,
    "SQLWARNING": tokens.Keyword,
    "STABLE": tokens.Keyword,
    "START": tokens.Keyword.DML,
    # 'STATE': tokens.Keyword,
    "STATEMENT": tokens.Keyword,
    "STATIC": tokens.Keyword,
    "STATISTICS": tokens.Keyword,
    "STDIN": tokens.Keyword,
    "STDOUT": tokens.Keyword,
    "STORAGE": tokens.Keyword,
    "STRICT": tokens.Keyword,
    "STRUCTURE": tokens.Keyword,
    "STYPE": tokens.Keyword,
    "SUBCLASS_ORIGIN": tokens.Keyword,
    "SUBLIST": tokens.Keyword,
    "SUBSTRING": tokens.Keyword,
    "SUCCESSFUL": tokens.Keyword,
    "SUM": tokens.Keyword,
    "SYMMETRIC": tokens.Keyword,
    "SYNONYM": tokens.Keyword,
    "SYSID": tokens.Keyword,
    "SYSTEM": tokens.Keyword,
    "SYSTEM_USER": tokens.Keyword,
    "TABLE": tokens.Keyword,
    "TABLE_NAME": tokens.Keyword,
    "TEMP": tokens.Keyword,
    "TEMPLATE": tokens.Keyword,
    "TEMPORARY": tokens.Keyword,
    "TERMINATE": tokens.Keyword,
    "THAN": tokens.Keyword,
    "TIMESTAMP": tokens.Keyword,
    "TIMEZONE_HOUR": tokens.Keyword,
    "TIMEZONE_MINUTE": tokens.Keyword,
    "TO": tokens.Keyword,
    "TOAST": tokens.Keyword,
    "TRAILING": tokens.Keyword,
    "TRANSATION": tokens.Keyword,
    "TRANSACTIONS_COMMITTED": tokens.Keyword,
    "TRANSACTIONS_ROLLED_BACK": tokens.Keyword,
    "TRANSATION_ACTIVE": tokens.Keyword,
    "TRANSFORM": tokens.Keyword,
    "TRANSFORMS": tokens.Keyword,
    "TRANSLATE": tokens.Keyword,
    "TRANSLATION": tokens.Keyword,
    "TREAT": tokens.Keyword,
    "TRIGGER": tokens.Keyword,
    "TRIGGER_CATALOG": tokens.Keyword,
    "TRIGGER_NAME": tokens.Keyword,
    "TRIGGER_SCHEMA": tokens.Keyword,
    "TRIM": tokens.Keyword,
    "TRUE": tokens.Keyword,
    "TRUNCATE": tokens.Keyword,
    "TRUSTED": tokens.Keyword,
    "TYPE": tokens.Keyword,
    "UID": tokens.Keyword,
    "UNCOMMITTED": tokens.Keyword,
    "UNDER": tokens.Keyword,
    "UNENCRYPTED": tokens.Keyword,
    "UNION": tokens.Keyword,
    "UNIQUE": tokens.Keyword,
    "UNKNOWN": tokens.Keyword,
    "UNLISTEN": tokens.Keyword,
    "UNNAMED": tokens.Keyword,
    "UNNEST": tokens.Keyword,
    "UNTIL": tokens.Keyword,
    "UPPER": tokens.Keyword,
    "USAGE": tokens.Keyword,
    "USE": tokens.Keyword,
    "USER": tokens.Keyword,
    "USER_DEFINED_TYPE_CATALOG": tokens.Keyword,
    "USER_DEFINED_TYPE_NAME": tokens.Keyword,
    "USER_DEFINED_TYPE_SCHEMA": tokens.Keyword,
    "USING": tokens.Keyword,
    "VACUUM": tokens.Keyword,
    "VALID": tokens.Keyword,
    "VALIDATE": tokens.Keyword,
    "VALIDATOR": tokens.Keyword,
    "VALUES": tokens.Keyword,
    "VARIABLE": tokens.Keyword,
    "VERBOSE": tokens.Keyword,
    "VERSION": tokens.Keyword,
    "VIEW": tokens.Keyword,
    "VOLATILE": tokens.Keyword,
    "WEEK": tokens.Keyword,
    "WHENEVER": tokens.Keyword,
    "WITH": tokens.Keyword.CTE,
    "WITHOUT": tokens.Keyword,
    "WORK": tokens.Keyword,
    "WRITE": tokens.Keyword,
    "YEAR": tokens.Keyword,
    "ZONE": tokens.Keyword,
    # Name.Builtin
    "ARRAY": tokens.Name.Builtin,
    "BIGINT": tokens.Name.Builtin,
    "BINARY": tokens.Name.Builtin,
    "BIT": tokens.Name.Builtin,
    "BLOB": tokens.Name.Builtin,
    "BOOLEAN": tokens.Name.Builtin,
    "CHAR": tokens.Name.Builtin,
    "CHARACTER": tokens.Name.Builtin,
    "DATE": tokens.Name.Builtin,
    "DEC": tokens.Name.Builtin,
    "DECIMAL": tokens.Name.Builtin,
    "FILE_TYPE": tokens.Name.Builtin,
    "FLOAT": tokens.Name.Builtin,
    "INT": tokens.Name.Builtin,
    "INT8": tokens.Name.Builtin,
    "INTEGER": tokens.Name.Builtin,
    "INTERVAL": tokens.Name.Builtin,
    "LONG": tokens.Name.Builtin,
    "NATURALN": tokens.Name.Builtin,
    "NVARCHAR": tokens.Name.Builtin,
    "NUMBER": tokens.Name.Builtin,
    "NUMERIC": tokens.Name.Builtin,
    "PLS_INTEGER": tokens.Name.Builtin,
    "POSITIVE": tokens.Name.Builtin,
    "POSITIVEN": tokens.Name.Builtin,
    "REAL": tokens.Name.Builtin,
    "ROWID": tokens.Name.Builtin,
    "ROWLABEL": tokens.Name.Builtin,
    "ROWNUM": tokens.Name.Builtin,
    "SERIAL": tokens.Name.Builtin,
    "SERIAL8": tokens.Name.Builtin,
    "SIGNED": tokens.Name.Builtin,
    "SIGNTYPE": tokens.Name.Builtin,
    "SIMPLE_DOUBLE": tokens.Name.Builtin,
    "SIMPLE_FLOAT": tokens.Name.Builtin,
    "SIMPLE_INTEGER": tokens.Name.Builtin,
    "SMALLINT": tokens.Name.Builtin,
    "SYS_REFCURSOR": tokens.Name.Builtin,
    "SYSDATE": tokens.Name,
    "TEXT": tokens.Name.Builtin,
    "TINYINT": tokens.Name.Builtin,
    "UNSIGNED": tokens.Name.Builtin,
    "UROWID": tokens.Name.Builtin,
    "UTL_FILE": tokens.Name.Builtin,
    "VARCHAR": tokens.Name.Builtin,
    "VARCHAR2": tokens.Name.Builtin,
    "VARYING": tokens.Name.Builtin,
}

KEYWORDS_COMMON = {
    "SELECT": tokens.Keyword.DML,
    "INSERT": tokens.Keyword.DML,
    "DELETE": tokens.Keyword.DML,
    "UPDATE": tokens.Keyword.DML,
    "UPSERT": tokens.Keyword.DML,
    "REPLACE": tokens.Keyword.DML,
    "MERGE": tokens.Keyword.DML,
    "DROP": tokens.Keyword.DDL,
    "CREATE": tokens.Keyword.DDL,
    "ALTER": tokens.Keyword.DDL,
    "WHERE": tokens.Keyword,
    "FROM": tokens.Keyword,
    "INNER": tokens.Keyword,
    "JOIN": tokens.Keyword,
    "STRAIGHT_JOIN": tokens.Keyword,
    "AND": tokens.Keyword,
    "OR": tokens.Keyword,
    "LIKE": tokens.Keyword,
    "ON": tokens.Keyword,
    "IN": tokens.Keyword,
    "SET": tokens.Keyword,
    "BY": tokens.Keyword,
    "GROUP": tokens.Keyword,
    "ORDER": tokens.Keyword,
    "LEFT": tokens.Keyword,
    "OUTER": tokens.Keyword,
    "FULL": tokens.Keyword,
    "IF": tokens.Keyword,
    "END": tokens.Keyword,
    "THEN": tokens.Keyword,
    "LOOP": tokens.Keyword,
    "AS": tokens.Keyword,
    "ELSE": tokens.Keyword,
    "FOR": tokens.Keyword,
    "WHILE": tokens.Keyword,
    "CASE": tokens.Keyword,
    "WHEN": tokens.Keyword,
    "MIN": tokens.Keyword,
    "MAX": tokens.Keyword,
    "DISTINCT": tokens.Keyword,
}

KEYWORDS_ORACLE = {
    "ARCHIVE": tokens.Keyword,
    "ARCHIVELOG": tokens.Keyword,
    "BACKUP": tokens.Keyword,
    "BECOME": tokens.Keyword,
    "BLOCK": tokens.Keyword,
    "BODY": tokens.Keyword,
    "CANCEL": tokens.Keyword,
    "CHANGE": tokens.Keyword,
    "COMPILE": tokens.Keyword,
    "CONTENTS": tokens.Keyword,
    "CONTROLFILE": tokens.Keyword,
    "DATAFILE": tokens.Keyword,
    "DBA": tokens.Keyword,
    "DISMOUNT": tokens.Keyword,
    "DOUBLE": tokens.Keyword,
    "DUMP": tokens.Keyword,
    "ELSIF": tokens.Keyword,
    "EVENTS": tokens.Keyword,
    "EXCEPTIONS": tokens.Keyword,
    "EXPLAIN": tokens.Keyword,
    "EXTENT": tokens.Keyword,
    "EXTERNALLY": tokens.Keyword,
    "FLUSH": tokens.Keyword,
    "FREELIST": tokens.Keyword,
    "FREELISTS": tokens.Keyword,
    # groups seems too common as table name
    # 'GROUPS': tokens.Keyword,
    "INDICATOR": tokens.Keyword,
    "INITRANS": tokens.Keyword,
    "INSTANCE": tokens.Keyword,
    "LAYER": tokens.Keyword,
    "LINK": tokens.Keyword,
    "LISTS": tokens.Keyword,
    "LOGFILE": tokens.Keyword,
    "MANAGE": tokens.Keyword,
    "MANUAL": tokens.Keyword,
    "MAXDATAFILES": tokens.Keyword,
    "MAXINSTANCES": tokens.Keyword,
    "MAXLOGFILES": tokens.Keyword,
    "MAXLOGHISTORY": tokens.Keyword,
    "MAXLOGMEMBERS": tokens.Keyword,
    "MAXTRANS": tokens.Keyword,
    "MINEXTENTS": tokens.Keyword,
    "MODULE": tokens.Keyword,
    "MOUNT": tokens.Keyword,
    "NOARCHIVELOG": tokens.Keyword,
    "NOCACHE": tokens.Keyword,
    "NOCYCLE": tokens.Keyword,
    "NOMAXVALUE": tokens.Keyword,
    "NOMINVALUE": tokens.Keyword,
    "NOORDER": tokens.Keyword,
    "NORESETLOGS": tokens.Keyword,
    "NORMAL": tokens.Keyword,
    "NOSORT": tokens.Keyword,
    "OPTIMAL": tokens.Keyword,
    "OWN": tokens.Keyword,
    "PACKAGE": tokens.Keyword,
    "PARALLEL": tokens.Keyword,
    "PCTINCREASE": tokens.Keyword,
    "PCTUSED": tokens.Keyword,
    "PLAN": tokens.Keyword,
    "PRIVATE": tokens.Keyword,
    "PROFILE": tokens.Keyword,
    "QUOTA": tokens.Keyword,
    "RECOVER": tokens.Keyword,
    "RESETLOGS": tokens.Keyword,
    "RESTRICTED": tokens.Keyword,
    "REUSE": tokens.Keyword,
    "ROLES": tokens.Keyword,
    "SAVEPOINT": tokens.Keyword,
    "SCN": tokens.Keyword,
    "SECTION": tokens.Keyword,
    "SEGMENT": tokens.Keyword,
    "SHARED": tokens.Keyword,
    "SNAPSHOT": tokens.Keyword,
    "SORT": tokens.Keyword,
    "STATEMENT_ID": tokens.Keyword,
    "STOP": tokens.Keyword,
    "SWITCH": tokens.Keyword,
    "TABLES": tokens.Keyword,
    "TABLESPACE": tokens.Keyword,
    "THREAD": tokens.Keyword,
    "TIME": tokens.Keyword,
    "TRACING": tokens.Keyword,
    "TRANSACTION": tokens.Keyword,
    "TRIGGERS": tokens.Keyword,
    "UNLIMITED": tokens.Keyword,
    "UNLOCK": tokens.Keyword,
}

# PostgreSQL Syntax
KEYWORDS_PLPGSQL = {
    "CONFLICT": tokens.Keyword,
    "WINDOW": tokens.Keyword,
    "PARTITION": tokens.Keyword,
    "OVER": tokens.Keyword,
    "PERFORM": tokens.Keyword,
    "NOTICE": tokens.Keyword,
    "PLPGSQL": tokens.Keyword,
    "INHERIT": tokens.Keyword,
    "INDEXES": tokens.Keyword,
    "ON_ERROR_STOP": tokens.Keyword,
    "BYTEA": tokens.Keyword,
    "BIGSERIAL": tokens.Keyword,
    "BIT VARYING": tokens.Keyword,
    "BOX": tokens.Keyword,
    "CHARACTER": tokens.Keyword,
    "CHARACTER VARYING": tokens.Keyword,
    "CIDR": tokens.Keyword,
    "CIRCLE": tokens.Keyword,
    "DOUBLE PRECISION": tokens.Keyword,
    "INET": tokens.Keyword,
    "JSON": tokens.Keyword,
    "JSONB": tokens.Keyword,
    "LINE": tokens.Keyword,
    "LSEG": tokens.Keyword,
    "MACADDR": tokens.Keyword,
    "MONEY": tokens.Keyword,
    "PATH": tokens.Keyword,
    "PG_LSN": tokens.Keyword,
    "POINT": tokens.Keyword,
    "POLYGON": tokens.Keyword,
    "SMALLSERIAL": tokens.Keyword,
    "TSQUERY": tokens.Keyword,
    "TSVECTOR": tokens.Keyword,
    "TXID_SNAPSHOT": tokens.Keyword,
    "UUID": tokens.Keyword,
    "XML": tokens.Keyword,
    "FOR": tokens.Keyword,
    "IN": tokens.Keyword,
    "LOOP": tokens.Keyword,
}

# Hive Syntax
KEYWORDS_HQL = {
    "EXPLODE": tokens.Keyword,
    "DIRECTORY": tokens.Keyword,
    "DISTRIBUTE": tokens.Keyword,
    "INCLUDE": tokens.Keyword,
    "LOCATE": tokens.Keyword,
    "OVERWRITE": tokens.Keyword,
    "POSEXPLODE": tokens.Keyword,
    "ARRAY_CONTAINS": tokens.Keyword,
    "CMP": tokens.Keyword,
    "COLLECT_LIST": tokens.Keyword,
    "CONCAT": tokens.Keyword,
    "CONDITION": tokens.Keyword,
    "DATE_ADD": tokens.Keyword,
    "DATE_SUB": tokens.Keyword,
    "DECODE": tokens.Keyword,
    "DBMS_OUTPUT": tokens.Keyword,
    "ELEMENTS": tokens.Keyword,
    "EXCHANGE": tokens.Keyword,
    "EXTENDED": tokens.Keyword,
    "FLOOR": tokens.Keyword,
    "FOLLOWING": tokens.Keyword,
    "FROM_UNIXTIME": tokens.Keyword,
    "FTP": tokens.Keyword,
    "HOUR": tokens.Keyword,
    "INLINE": tokens.Keyword,
    "INSTR": tokens.Keyword,
    "LEN": tokens.Keyword,
    "MAP": tokens.Name.Builtin,
    "MAXELEMENT": tokens.Keyword,
    "MAXINDEX": tokens.Keyword,
    "MAX_PART_DATE": tokens.Keyword,
    "MAX_PART_INT": tokens.Keyword,
    "MAX_PART_STRING": tokens.Keyword,
    "MINELEMENT": tokens.Keyword,
    "MININDEX": tokens.Keyword,
    "MIN_PART_DATE": tokens.Keyword,
    "MIN_PART_INT": tokens.Keyword,
    "MIN_PART_STRING": tokens.Keyword,
    "NOW": tokens.Keyword,
    "NVL": tokens.Keyword,
    "NVL2": tokens.Keyword,
    "PARSE_URL_TUPLE": tokens.Keyword,
    "PART_LOC": tokens.Keyword,
    "PART_COUNT": tokens.Keyword,
    "PART_COUNT_BY": tokens.Keyword,
    "PRINT": tokens.Keyword,
    "PUT_LINE": tokens.Keyword,
    "RANGE": tokens.Keyword,
    "REDUCE": tokens.Keyword,
    "REGEXP_REPLACE": tokens.Keyword,
    "RESIGNAL": tokens.Keyword,
    "RTRIM": tokens.Keyword,
    "SIGN": tokens.Keyword,
    "SIGNAL": tokens.Keyword,
    "SIN": tokens.Keyword,
    "SPLIT": tokens.Keyword,
    "SQRT": tokens.Keyword,
    "STACK": tokens.Keyword,
    "STR": tokens.Keyword,
    "STRING": tokens.Name.Builtin,
    "STRUCT": tokens.Name.Builtin,
    "SUBSTR": tokens.Keyword,
    "SUMMARY": tokens.Keyword,
    "TBLPROPERTIES": tokens.Keyword,
    "TIMESTAMP": tokens.Name.Builtin,
    "TIMESTAMP_ISO": tokens.Keyword,
    "TO_CHAR": tokens.Keyword,
    "TO_DATE": tokens.Keyword,
    "TO_TIMESTAMP": tokens.Keyword,
    "TRUNC": tokens.Keyword,
    "UNBOUNDED": tokens.Keyword,
    "UNIQUEJOIN": tokens.Keyword,
    "UNIX_TIMESTAMP": tokens.Keyword,
    "UTC_TIMESTAMP": tokens.Keyword,
    "VIEWS": tokens.Keyword,
    "EXIT": tokens.Keyword,
    "BREAK": tokens.Keyword,
    "LEAVE": tokens.Keyword,
}


KEYWORDS_MSACCESS = {
    "DISTINCTROW": tokens.Keyword,
}



import unittest

from pyspark.sql import DataFrame
from pyspark.sql.types import StringType, StructField, StructType

from spetlr.functions import get_unique_tempview_name
from spetlr.spark import Spark
from spetlr.transformations import merge_df_into_target
from spetlr.utils import DataframeCreator


class MergeDfIntoTargetTest(unittest.TestCase):
    db_name = "test" + get_unique_tempview_name()
    table_name = "testTarget"

    schema = StructType(
        [
            StructField("Id", StringType(), False),
            StructField("Brand", StringType(), True),
            StructField("Model", StringType(), True),
        ]
    )

    cols = ["Id", "Brand", "Model"]

    row1 = ("1", "Fender", "Jaguar")
    row2 = ("2", "Gibson", "Starfire")
    row3 = ("3", "Ibanez", "RG")
    data_rows = [
        row1,
        row2,
        row3,
    ]
    targetrow1 = ("0", "Fender", "Telecaster")
    targetrow2 = ("2", "Gibson", "Les Paul")

    @classmethod
    def setUpClass(cls):
        cls.create_database(cls.db_name)
        cls.create_test_table(cls.table_name, cls.db_name)

    @classmethod
    def tearDownClass(cls) -> None:
        Spark.get().sql(f"drop database {cls.db_name} cascade")

    def test_01_insert(self):
        """Tests that a new row is inserted

        test table before:
        +----+-----+----+----------------+
        |Id  |    Brand |           Model|
        +----+-----+----+----------------+
        |   0|    Fender|      Telecaster|
        +----+----------+---------------+

        test table after:
        +----+-----+----+----------------+
        |Id  |    Brand |           Model|
        +----+-----+----+----------------+
        |   0|    Fender|     Telecaster|
        |   1|    Fender|          Jaguar|
        |   2|    Gibson|        Starfire|
        |   3|    Ibanez|              RG|
        +----+----------+---------------+

        """
        # Truncate table
        Spark.get().sql(f"truncate table {self.db_name}.{self.table_name}")

        # Create target data
        Spark.get().sql(
            f"INSERT INTO {self.db_name}.{self.table_name} values {self.targetrow1}"
        )

        #  Merge dataframe into target
        df = self.create_data()
        merge_df_into_target(df, self.table_name, self.db_name, ["Id"])

        # Compare
        df_expected = self.expected_data_01()
        df_result = self.get_target_table()
        self.equal_dfs(df_expected, df_result)

    def test_02_merge(self):
        """Tests that a new row is merged

         test table before:
        +----+-----+----+----------------+
        |Id  |    Brand |           Model|
        +----+-----+----+----------------+
        |   2|    Gibson|      Les Paul|
        +----+----------+---------------+

        test table after:
        +----+-----+----+----------------+
        |Id  |    Brand |           Model|
        +----+-----+----+----------------+
        |   1|    Fender|          Jaguar|
        |   2|    Gibson|        Starfire|
        |   3|    Ibanez|              RG|
        +----+----------+---------------+
        """

        # Truncate table
        Spark.get().sql(f"truncate table {self.db_name}.{self.table_name}")

        # Create target data
        Spark.get().sql(
            f"INSERT INTO {self.db_name}.{self.table_name} values {self.targetrow2}"
        )

        #  Merge dataframe into target
        df = self.create_data()
        merge_df_into_target(df, self.table_name, self.db_name, ["Id"])

        # Compare
        df_expected = self.expected_data_02()
        df_result = self.get_target_table()
        self.equal_dfs(df_expected, df_result)

    def test_03_merge_insert(self):
        """Tests that one row is merged and one inserted

         test table before:
        +----+-----+----+----------------+
        |Id  |    Brand |           Model|
        +----+-----+----+----------------+
        |   0|    Fender|      Telecaster|
        |   2|    Gibson|      Les Paul|
        +----+----------+---------------+

        test table after:
        +----+-----+----+----------------+
        |Id  |    Brand |           Model|
        +----+-----+----+----------------+
        |   0|    Fender|      Telecaster|
        |   1|    Fender|          Jaguar|
        |   2|    Gibson|        Starfire|
        |   3|    Ibanez|              RG|
        +----+----------+---------------+
        """

        # Truncate table
        Spark.get().sql(f"truncate table {self.db_name}.{self.table_name}")

        # Create target data
        Spark.get().sql(
            f"INSERT INTO {self.db_name}.{self.table_name} values {self.targetrow1}"
        )
        Spark.get().sql(
            f"INSERT INTO {self.db_name}.{self.table_name} values {self.targetrow2}"
        )

        #  Merge dataframe into target
        df = self.create_data()
        merge_df_into_target(df, self.table_name, self.db_name, ["Id"])

        # Compare
        df_expected = self.expected_data_03()
        df_result = self.get_target_table()
        self.equal_dfs(df_expected, df_result)

    @classmethod
    def create_test_table(self, table_name="testTarget", db_name="test"):
        location = f"/tmp/{db_name}/{table_name}"
        sql_argument = f"""CREATE TABLE IF NOT EXISTS {db_name}.{table_name}(
                      Id STRING,
                      Brand STRING,
                      Model STRING
                      )
                      USING DELTA
                      LOCATION '{location}'"""
        Spark.get().sql(sql_argument)

    @classmethod
    def create_database(self, db_name="test") -> None:
        location = f"/tmp/{db_name}/"
        sql_argument = f"CREATE DATABASE IF NOT EXISTS {db_name} LOCATION '{location}'"
        Spark.get().sql(sql_argument)

    def create_data(self) -> DataFrame:
        df_new = DataframeCreator.make_partial(
            schema=self.schema,
            columns=self.cols,
            data=self.data_rows,
        )

        return df_new.orderBy("Id")

    def expected_data_01(self) -> DataFrame:
        df_new = DataframeCreator.make_partial(
            schema=self.schema,
            columns=self.cols,
            data=[self.targetrow1, self.row1, self.row2, self.row3],
        )

        return df_new.orderBy("Id")

    def expected_data_02(self) -> DataFrame:
        df_new = DataframeCreator.make_partial(
            schema=self.schema,
            columns=self.cols,
            data=[self.row1, self.row2, self.row3],
        )

        return df_new.orderBy("Id")

    def expected_data_03(self) -> DataFrame:
        df_new = DataframeCreator.make_partial(
            schema=self.schema,
            columns=self.cols,
            data=[self.targetrow1, self.row1, self.row2, self.row3],
        )

        return df_new.orderBy("Id")

    def get_target_table(self):
        return Spark.get().read.table(f"{self.db_name}.{self.table_name}").orderBy("Id")

    def equal_dfs(self, df1, df2):
        df_expected_pd = df1.toPandas()
        df_result_pd = df2.toPandas()

        self.assertTrue(df_result_pd.equals(df_expected_pd))



import unittest

from spetlr.sql import SqlExecutor
from tests.local.sql import sql


class TestSqlExecutor(unittest.TestCase):
    def test_01_standard_splitting(self):
        s = SqlExecutor(sql)
        statements = list(s.get_statements("*"))
        self.assertEqual(len(statements), 3)

    def test_02_marker_splitting(self):
        s = SqlExecutor(sql, statement_spliter=["-- COMMAND ----------"])
        statements = list(s.get_statements("*"))
        self.assertEqual(len(statements), 2)

    def test_03_no_splitting(self):
        s = SqlExecutor(sql, statement_spliter=None)
        statements = list(s.get_statements("*"))
        self.assertEqual(len(statements), 1)



from typing import Protocol

from pyspark.sql import DataFrame

from spetlr.tables import TableHandle


class CommonBaseServer(Protocol):
    def execute_sql(self, sql: str) -> None:
        """synonym of .sql()"""
        pass

    def from_tc(self, id: str) -> TableHandle:
        pass

    def sql(self, sql: str) -> None:
        """synonym of .execute_sql()"""
        pass

    def read_table_by_name(self, table_name: str) -> DataFrame:
        pass

    def write_table_by_name(
        self,
        df_source: DataFrame,
        table_name: str,
        append: bool = False,
    ):
        pass



import pyspark.sql.functions as F
from pyspark.sql import DataFrame

from spetlr.etl import Transformer


class DataFrameFilterTransformer(Transformer):
    def __init__(self, col_value: str, col_name: str = "messageType") -> None:
        """
        A simple helper class to filter a dataframe based on the passed col_name
        if its value equals to the specified col_value.

        Args:
            col_value: The value of the column to filter the dataframe
            col_name: The name of the column to filter the dataframe. The default
                is 'messageType' as this class is mostly used to filter dataframes
                with this column name.
        """
        super().__init__()
        self.col_value = col_value
        self.col_name = col_name

    def process(self, df: DataFrame) -> DataFrame:
        """
        Method to filter the passed dataframe df based on the col_name and col_value
        of the class.

        Args:
            df: The dataframe to filter

        Returns:
            The filtered dataframe
        """
        return df.filter(F.col(self.col_name) == self.col_value)



import difflib
from typing import Dict, Iterable, List

from pyspark.sql import DataFrame

from spetlr.etl.transformer import Transformer
from spetlr.exceptions import SpetlrException


class FuzzySelectException(SpetlrException):
    pass


class NoMatchException(FuzzySelectException):
    pass


class NonUniqueException(SpetlrException):
    pass


class FuzzySelectTransformer(Transformer):
    """To construct a FuzzySelectTransformer,
    give it a list of columns that you want to see in the output.

    The fuzzy aspect is that difflib is used to find the best
    approximate source column for each column name. The returned
    df will contain the columns as given to this transformer.

    If you would like to check the set of columns that your
    dataframe will be transformed into you can directly call
    the method find_best_mapping and inspect the returned mapping.
    """

    def __init__(self, columns: Iterable[str], match_cutoff=0.6):
        super().__init__()
        self.columns = list(columns)
        self.match_cutoff = match_cutoff

    def find_best_mapping(self, in_columns: Iterable[str]) -> Dict[str, str]:
        """returns a dict mapping source to target columns
        as a case insensitive fuzzy match."""
        in_columns = {col.lower(): col for col in in_columns}
        mapping = dict()
        for col in self.columns:
            if not len(in_columns):
                NoMatchException("Not enough source columns to match from.")

            matches: List[str] = difflib.get_close_matches(
                col.lower(), list(in_columns.keys()), n=2, cutoff=self.match_cutoff
            )

            if not len(matches):
                raise NoMatchException(f"No suitable source column found for {col}")

            if len(matches) > 1:
                candidates = [
                    (
                        in_columns[match],
                        difflib.SequenceMatcher(a=col.lower(), b=match).ratio(),
                    )
                    for match in matches
                ]
                raise NonUniqueException(
                    f"No unique source column match for {col}: candidates: {candidates}"
                )

            source_match = in_columns[matches[0]]
            mapping[source_match] = col
            del in_columns[matches[0]]

        return mapping

    def process(self, df: DataFrame) -> DataFrame:
        mapping = self.find_best_mapping(df.columns)

        new_df = df
        for key, value in mapping.items():
            new_df = new_df.withColumnRenamed(key, value)

        new_df = new_df.select(*self.columns)
        return new_df



from typing import List, Union

import pyspark.sql.types as T
from pyspark.sql import DataFrame

from spetlr.etl import TransformerNC
from spetlr.utils import SelectAndCastColumns


class SelectAndCastColumnsTransformerNC(TransformerNC):
    """
    This transformer select and cast columns based pyspark schema.

    Attributes:
    ----------
        schema : StructType
            pyspark schema to use for select and cast
        caseInsensitiveMatching : Boolean
            used for case insensitive matching, by default False
        dataset_input_keys : Union[str, List[str]]
            list of input dataset keys
        dataset_output_key : str
            output dataset key
    """

    def __init__(
        self,
        *,
        schema: T.StructType,
        caseInsensitiveMatching: bool = False,
        dataset_input_keys: Union[str, List[str]] = None,
        dataset_output_key: str = None,
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
        )
        self.schema = schema
        self.caseInsensitiveMatching = caseInsensitiveMatching

    def process(self, df: DataFrame) -> DataFrame:
        return SelectAndCastColumns(
            df=df,
            schema=self.schema,
            caseInsensitiveMatching=self.caseInsensitiveMatching,
        )



from typing import Dict

from more_itertools import peekable

from spetlr.configurator.sql import sqlparse
from spetlr.configurator.sql.db import _walk_db_statement
from spetlr.configurator.sql.table import _walk_table_statement
from spetlr.configurator.sql.utils import _meaningful_token_iter


def _walk_create_statement(statement: sqlparse.sql.Statement) -> Dict:
    # create an iterator over all tokens in the statement
    # that are not whitespace
    statement = peekable(_meaningful_token_iter(statement))

    # We are now ready to go through the statement
    # both create Table and create database must start with create
    token = next(statement)
    if token.value.upper() != "CREATE":
        # not a table or db create statement. Nothing to do
        return {}

    entity = next(statement)

    if entity.value.upper() == "TABLE":
        is_table = True
    elif entity.value.upper() in ("SCHEMA", "DATABASE"):
        is_table = False
    else:
        # not a table or db create statement. Nothing to do
        return {}

    # next is the optional "IF NOT EXISTS"
    if statement.peek().value.upper() == "IF NOT EXISTS":
        next(statement)

    if is_table:
        return _walk_table_statement(statement)
    else:
        return _walk_db_statement(statement)



from pyspark.sql import types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.spark import Spark
from spetlr.transformers import TimeZoneTransformerNC


class TimeZoneTransformerNCTest(DataframeTestCase):
    def test_timezone_transformer(self):
        input_schema = T.StructType(
            [
                T.StructField("Latitude", T.DoubleType(), True),
                T.StructField("Longitude", T.DoubleType(), True),
            ]
        )

        input_data = [(51.519487, -0.083069), (55.6761, 12.5683)]

        input_df = Spark.get().createDataFrame(data=input_data, schema=input_schema)

        transformed_df = TimeZoneTransformerNC(
            latitude_col="Latitude",
            longitude_col="Longitude",
        ).process(input_df)

        expected_data = [
            (51.519487, -0.083069, "Europe/London"),
            (55.6761, 12.5683, "Europe/Copenhagen"),
        ]

        self.assertDataframeMatches(
            df=transformed_df,
            columns=None,
            expected_data=expected_data,
        )

    def test_timezone_transformer_none(self):
        input_schema = T.StructType(
            [
                T.StructField("Latitude", T.DoubleType(), True),
                T.StructField("Longitude", T.DoubleType(), True),
            ]
        )

        input_data = [(None, -0.083069), (55.6761, None), (None, None)]

        input_df = Spark.get().createDataFrame(data=input_data, schema=input_schema)

        transformed_df = TimeZoneTransformerNC(
            latitude_col="Latitude",
            longitude_col="Longitude",
        ).process(input_df)

        transformed_data = [tuple(row) for row in transformed_df.collect()]

        expected_data = [
            (None, -0.083069, None),
            (55.6761, None, None),
            (None, None, None),
        ]

        self.assertEqual(
            first=transformed_data,
            second=expected_data,
        )



from deprecated import deprecated

from spetlr.sql.SqlServer import SqlServer

from . import SimpleLoader


@deprecated(reason="Use a SimpleLoader with a SqlHandle.")
class SimpleSqlServerLoader(SimpleLoader):
    def __init__(
        self,
        *,
        table_id: str,
        server: SqlServer,
        append: bool = False,
    ):
        super().__init__(
            handle=server.from_tc(table_id), mode="append" if append else "overwrite"
        )



import subprocess
import sys
import unittest
from pathlib import Path


class OrchestratorTests(unittest.TestCase):
    def test_examples(self):
        examples_folder = (
            Path(__file__).parent.parent.parent.parent / "examples" / "etl"
        )
        for file in examples_folder.iterdir():
            if file.suffix == ".py":
                subprocess.run([sys.executable, str(file)], check=True)


if __name__ == "__main__":
    unittest.main()



from datetime import timedelta
from typing import List

from pyspark.sql.types import (
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from spetlrtools.testing import DataframeTestCase, TestHandle
from spetlrtools.time import dt_utc

from spetlr.etl.extractors.incremental_extractor import IncrementalExtractor
from spetlr.utils import DataframeCreator


class IncrementalExtractorTests(DataframeTestCase):
    date_row1 = dt_utc(2021, 1, 1, 10, 50)  # 1st of january 2021, 10:50
    date_row2 = dt_utc(2021, 1, 1, 10, 55)  # 1st of january 2021, 10:55
    date_row2Inc = dt_utc(2021, 1, 1, 10, 56)  # 1st of january 2021, 10:56
    date_row3 = dt_utc(2021, 1, 1, 11, 00)  # 1st of january 2021, 11:00
    date_row4 = dt_utc(2021, 1, 5, 11, 00)  # 5th of january 2021, 11:00

    row1 = (1, "string1", date_row1)
    row2 = (2, "string2", date_row2)
    row2Inc = (22, "string2Inc", date_row2Inc)
    row3 = (3, "String3", date_row3)
    row4 = (4, "String4", date_row4)

    # Test data for empty target or source
    source1 = [row1, row2, row3]
    target1 = [row1, row2]

    # Test data for incremental extraction
    source1Inc = [row1, row2Inc, row3]
    target1Inc = [row1, row2]
    extract1Inc = [row2Inc, row3]

    # Test data for overlapping period
    sourceOverlap = [row2, row3, row4]
    targetOverlap = [row2, row3, row4]
    extractOverlap = [row2, row3, row4]

    # The columns of the tables
    dummy_columns: List[str] = ["id", "stringcol", "timecol"]

    # The schema of the tables
    dummy_schema = StructType(
        [
            StructField("id", IntegerType(), True),
            StructField("stringcol", StringType(), True),
            StructField("timecol", TimestampType(), True),
        ]
    )

    def test_01_can_perform_incremental_from_empty_source(self):
        """Source is empty. Target has data. No data is read.


        Source has no data:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |  |              |                  |

        Target has the following data:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |1 | "string1"    | 01.01.2021 10:50 |
        |2| "string2"     | 01.01.2021 10:55 |

        So an empty dataframe is read:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |  |              |                  |

        """

        source_test_handle = TestHandle(
            provides=DataframeCreator.make_partial(
                self.dummy_schema, self.dummy_columns, []
            )
        )

        target_test_handle = TestHandle(
            provides=DataframeCreator.make_partial(
                self.dummy_schema, self.dummy_columns, self.target1
            )
        )

        extractor = IncrementalExtractor(
            handle_source=source_test_handle,
            handle_target=target_test_handle,
            time_col_source="timecol",
            time_col_target="timecol",
            dataset_key="source",
        )

        df_result = extractor.read()

        self.assertDataframeMatches(df_result, None, [])

    def test_02_can_perform_incremental_on_empty_target(self):
        """Source has data. Target is empty. All source data are read.

        Source has the following data:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |1 | "string1"    | 01.01.2021 10:50 |
        |2| "string2"     | 01.01.2021 10:55 |
        |3 | "string3"    | 01.01.2021 11:00 |

        Target has no data:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |  |              |                  |

        The following data (all source data) is read:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |1 | "string1"    | 01.01.2021 10:50 |
        |2| "string2"     | 01.01.2021 10:55 |
        |3 | "string3"    | 01.01.2021 11:00 |
        """
        source_test_handle = TestHandle(
            provides=DataframeCreator.make_partial(
                self.dummy_schema, self.dummy_columns, self.source1
            )
        )

        target_test_handle = TestHandle(
            provides=DataframeCreator.make_partial(
                self.dummy_schema, self.dummy_columns, []
            )
        )

        extractor = IncrementalExtractor(
            handle_source=source_test_handle,
            handle_target=target_test_handle,
            time_col_source="timecol",
            time_col_target="timecol",
            dataset_key="source",
        )

        df_extract = extractor.read()

        self.assertDataframeMatches(df_extract, None, self.source1)

    def test_03_can_extract_incremental(self):
        """
        Source has the following data:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |1 | "string1"    | 01.01.2021 10:50 |
        |22| "string2inc" | 01.01.2021 10:56 |
        |3 | "string3"    | 01.01.2021 11:00 |


        Target has the following data:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |1 | "string1"    | 01.01.2021 10:50 |
        |2| "string2"     | 01.01.2021 10:55 |

        So data from after 01.01.2021 10:55 should be read:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |22| "string2inc" | 01.01.2021 10:56 |
        |3 | "string3"    | 01.01.2021 11:00 |
        """
        source_test_handle = TestHandle(
            provides=DataframeCreator.make_partial(
                self.dummy_schema, self.dummy_columns, self.source1Inc
            )
        )

        target_test_handle = TestHandle(
            provides=DataframeCreator.make_partial(
                self.dummy_schema, self.dummy_columns, self.target1Inc
            )
        )

        extractor = IncrementalExtractor(
            handle_source=source_test_handle,
            handle_target=target_test_handle,
            time_col_source="timecol",
            time_col_target="timecol",
            dataset_key="source",
        )

        df_extract = extractor.read()

        self.assertDataframeMatches(df_extract, None, self.extract1Inc)

    def test_04_can_extract_with_overlap(self):
        """
        Source has the following data:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |2 | "string2"    | 01.01.2021 10:55 |
        |3|  "string3"    | 01.01.2021 11:00 |
        |4 | "string4"    | 05.01.2021 11:00 |


        Target has the following data:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |2 | "string2"    | 01.01.2021 10:55 |
        |3|  "string3"    | 01.01.2021 11:00 |
        |4 | "string4"    | 05.01.2021 11:00 |

        Since an overlap of 5 days id defined all data should have been read:

        |id| stringcol    | timecol          |
        |--|--------------|------------------|
        |2 | "string2"    | 01.01.2021 10:55 |
        |3|  "string3"    | 01.01.2021 11:00 |
        |4 | "string4"    | 05.01.2021 11:00 |
        """
        source_test_handle = TestHandle(
            provides=DataframeCreator.make_partial(
                self.dummy_schema, self.dummy_columns, self.sourceOverlap
            )
        )

        target_test_handle = TestHandle(
            provides=DataframeCreator.make_partial(
                self.dummy_schema, self.dummy_columns, self.targetOverlap
            )
        )

        extractor = IncrementalExtractor(
            handle_source=source_test_handle,
            handle_target=target_test_handle,
            time_col_source="timecol",
            time_col_target="timecol",
            dataset_key="source",
            overlap_period=timedelta(days=5),
        )

        df_extract = extractor.read()

        self.assertDataframeMatches(df_extract, None, self.extractOverlap)



import argparse

from spetlr.exceptions.cli_exceptions import SpetlrCliException

from . import generate_keys_file


def cli(self):
    """
    This function exposes functionality to be used in code analysis and deployment.
    Initialize the configurator in your project like

    ```
    def init_my_configurator():
        c= Configurator()
        c.add_resource_path(my_yaml_module)
        c.register('ENV', config.my_env_name)
        return c

    def my_configurator_cli():
        init_my_configurator().cli()
    ```
    Then expose the `my_configurator_cli` as a command line tool.
    The cli has the following usage

    <my_cli> [action [action_options]]

    Available actions:

    - generate-keys-file [-o output_file]
        Generates a python file that contains every available key string as a python
        object of the same name. Including this file in your python project allows
        you to get auto-completion of key names.
        The output is dumped to stdout unless -o output_file is specified.
        The option -o has the side-effect of return an exit code 1 if the file was
        updated. This allows you to check for a correctly updated keys file in your
        CICD pipleine.
    """
    parser = argparse.ArgumentParser()
    subp = parser.add_subparsers(help="actions")

    generate_keys_file.setup_parser(subp.add_parser("generate-keys-file"))
    # add further parsers here

    options = parser.parse_args()

    try:
        func = options.func
    except AttributeError:
        parser.print_help()
        raise SpetlrCliException("No actions given")

    func(self, options)



from abc import ABC
from typing import Any, List, Union

from pyspark.sql import DataFrame


class TableHandle(ABC):
    def read(self) -> DataFrame:
        pass

    def overwrite(self, df: DataFrame) -> None:
        pass

    def append(self, df: DataFrame) -> None:
        pass

    def truncate(self) -> None:
        pass

    def drop(self) -> None:
        pass

    def drop_and_delete(self) -> None:
        pass

    def write_or_append(self, df: DataFrame, mode: str) -> None:
        pass

    def upsert(self, df: DataFrame, join_cols: List[str]) -> Union[DataFrame, None]:
        pass

    def get_tablename(self) -> str:
        pass

    def delete_data(
        self, comparison_col: str, comparison_limit: Any, comparison_operator: str
    ) -> None:
        pass



import random
import unittest
from unittest.mock import create_autospec

from pyspark.sql.types import IntegerType, StructField, StructType
from spetlrtools.testing.DataframeTestCase import DataframeTestCase

from spetlr.functions import get_unique_tempview_name
from spetlr.spark import Spark
from spetlr.sql import SqlHandle
from spetlr.sql.sql_handle import SqlHandleInvalidName
from spetlr.sql.SqlBaseServer import SqlBaseServer


class SqlHandleTests(DataframeTestCase):
    sql_server_mock = None
    table_handle = None
    test_table_name = get_unique_tempview_name()
    test_join_cols = ["testcolumn"]
    test_df_data = [(random.getrandbits(16),)]
    test_df_schema = StructType(
        [
            StructField("testcolumn", IntegerType(), True),
        ]
    )
    test_df = Spark.get().createDataFrame(data=test_df_data, schema=test_df_schema)

    def setUp(self):
        self.sql_server_mock = create_autospec(SqlBaseServer)
        self.table_handle = SqlHandle(
            name=self.test_table_name, sql_server=self.sql_server_mock
        )

    def test_01_validate_with_table_name(self):
        _ = SqlHandle(
            name="tablename",
            sql_server=self.sql_server_mock,
        )
        self.assertTrue(True)

    def test_02_validate_with_db_name_and_table_name(self):
        _ = SqlHandle(
            name="dbname.tablename",
            sql_server=self.sql_server_mock,
        )
        self.assertTrue(True)

    def test_03_not_valid_name(self):
        with self.assertRaises(SqlHandleInvalidName):
            _ = SqlHandle(
                name="dbname.tablename.test",
                sql_server=self.sql_server_mock,
            )

        self.assertTrue(True)

    def test_04_read(self):
        self.sql_server_mock.read_table_by_name.return_value = self.test_df

        df_return = self.table_handle.read()
        self.assertDataframeMatches(df=df_return, expected_data=self.test_df_data)

        self.assertEqual(self.sql_server_mock.read_table_by_name.call_count, 1)
        args = self.sql_server_mock.read_table_by_name.call_args[1]
        self.assertEqual(args["table_name"], self.test_table_name)

    def test_04_overwrite(self):
        self.table_handle.overwrite(df=self.test_df)

        self.assertEqual(self.sql_server_mock.write_table_by_name.call_count, 1)
        args = self.sql_server_mock.write_table_by_name.call_args[1]

        self.assertDataframeMatches(
            df=args["df_source"], expected_data=self.test_df_data
        )
        self.assertEqual(args["table_name"], self.test_table_name)
        self.assertFalse(args["append"])

    def test_05_append(self):
        self.table_handle.append(df=self.test_df)

        self.assertEqual(self.sql_server_mock.write_table_by_name.call_count, 1)
        args = self.sql_server_mock.write_table_by_name.call_args[1]

        self.assertDataframeMatches(
            df=args["df_source"], expected_data=self.test_df_data
        )
        self.assertEqual(args["table_name"], self.test_table_name)
        self.assertTrue(args["append"])

    def test_06_write_or_append_overwrite(self):
        self.table_handle.write_or_append(df=self.test_df, mode="overwrite")

        self.assertEqual(self.sql_server_mock.write_table_by_name.call_count, 1)
        args = self.sql_server_mock.write_table_by_name.call_args[1]

        self.assertDataframeMatches(
            df=args["df_source"], expected_data=self.test_df_data
        )
        self.assertEqual(args["table_name"], self.test_table_name)
        self.assertFalse(args["append"])

    def test_07_write_or_append_append(self):
        self.table_handle.write_or_append(df=self.test_df, mode="append")

        self.assertEqual(self.sql_server_mock.write_table_by_name.call_count, 1)
        args = self.sql_server_mock.write_table_by_name.call_args[1]

        self.assertDataframeMatches(
            df=args["df_source"], expected_data=self.test_df_data
        )
        self.assertEqual(args["table_name"], self.test_table_name)
        self.assertTrue(args["append"])

    def test_07_upsert(self):
        self.table_handle.upsert(df=self.test_df, join_cols=self.test_join_cols)

        self.assertEqual(self.sql_server_mock.upsert_to_table_by_name.call_count, 1)
        args = self.sql_server_mock.upsert_to_table_by_name.call_args[1]

        self.assertDataframeMatches(
            df=args["df_source"], expected_data=self.test_df_data
        )
        self.assertEqual(args["table_name"], self.test_table_name)
        self.assertEqual(args["join_cols"], self.test_join_cols)

    def test_08_truncate(self):
        self.table_handle.truncate()

        self.assertEqual(self.sql_server_mock.truncate_table_by_name.call_count, 1)
        args = self.sql_server_mock.truncate_table_by_name.call_args[1]

        self.assertEqual(args["table_name"], self.test_table_name)

    def test_09_drop(self):
        self.table_handle.drop()

        self.assertEqual(self.sql_server_mock.drop_table_by_name.call_count, 1)
        args = self.sql_server_mock.drop_table_by_name.call_args[1]

        self.assertEqual(args["table_name"], self.test_table_name)

    def test_10_drop_and_delete(self):
        self.table_handle.drop_and_delete()

        self.assertEqual(self.sql_server_mock.drop_table_by_name.call_count, 1)
        args = self.sql_server_mock.drop_table_by_name.call_args[1]

        self.assertEqual(args["table_name"], self.test_table_name)


if __name__ == "__main__":
    unittest.main()



import unittest

import pyspark.sql.types as T

from spetlr.configurator import Configurator
from spetlr.schema_manager import SchemaManager

from . import extras
from .extras import initSchemaManager


class TestSchemaManager(unittest.TestCase):
    sc: SchemaManager

    @classmethod
    def setUpClass(cls) -> None:
        c = Configurator()
        c.clear_all_configurations()
        c.add_resource_path(extras)
        SchemaManager().clear_all_configurations()
        cls.sc = initSchemaManager()

    def test_register_schema(self):
        schema = T.StructType(
            [
                T.StructField("Column1", T.IntegerType(), True),
                T.StructField("Column2", T.StringType(), True),
                T.StructField("Column3", T.FloatType(), True),
            ]
        )

        self.assertNotIn("register_test", self.sc._registered_schemas)

        self.sc.register_schema(schema_name="register_test", schema=schema)

        self.assertIn("register_test", self.sc._registered_schemas)

    def test_get_registered_schema(self):
        schema = T.StructType(
            [
                T.StructField("Column1", T.IntegerType(), True),
                T.StructField("Column2", T.StringType(), True),
                T.StructField("Column3", T.FloatType(), True),
            ]
        )

        self.assertNotIn("register_test2", self.sc._registered_schemas)

        self.sc.register_schema(schema_name="register_test2", schema=schema)

        self.assertEqual(
            schema,
            self.sc.get_schema(schema_identifier="register_test2"),
        )

    def test_get_python_ref_schema(self):
        schema = SchemaManager().get_schema(schema_identifier="SchemaTestTable1")

        expected_schema = extras.python_test_schema

        self.assertEqual(expected_schema, schema)



import re
from typing import Optional

from more_itertools import peekable
from pyspark.sql import types as t

from spetlr.configurator.sql.init_sqlparse import parse
from spetlr.configurator.sql.utils import _meaningful_token_iter


class SchemaExtractionError(Exception):
    pass


def get_schema(sql: str) -> t.StructType:
    """from sql table schema to structType"""
    sql = sql.strip()
    if sql[0] == "(":
        sql = sql[1:-1]
    parsed_schema = parse(sql)
    if not len(parsed_schema) == 1:
        raise SchemaExtractionError("multiple statements")
    (tokenlist,) = parsed_schema

    iter = peekable(_meaningful_token_iter(tokenlist.flatten()))

    return _get_schema(iter)


# throughout this file, the variable iter represents the following object:
# - a peekable iterator
# - over a flattened list of sql tokens
# - where comments and whitespaces have been removed.
# as an iterator it has a state and will therefore refer to some local position
# of a stream at the start of the function. After the function call, it will point
# to the position after the end of the processed tokens (if any)
# any call to next or peek can throw StopIteration, so be sure to always maintain
# a coherent state in case of unexpected function exit.
#
# the underscore methods are private and no TypeHint is used. If you can capture the
# above in a type-hint, please consider a contribution.
def _get_schema(iter) -> t.StructType:
    """Parse everything in the iterator as a struct
    will parse with or without ":" between name and type and will parse
    until the closing > or end of string is detected.
    The closing ">" will be consumed.
    """
    struct = t.StructType()
    # parse the variables one by one
    for token in iter:
        # the first token is taken to be the name.
        # since almost anything can be a name, no validation is done.
        name = token.value.strip("`")

        colon_token = iter.peek()
        if colon_token.value == ":":
            next(iter)  # ignore the colon

        # get the data type including any definitions for struct, arrays, maps
        dataType = _get_data_type(iter)

        # create the member now and modify it further as needed
        # this simplifies early return
        struct.add(name, dataType)

        # try parsing any further modifiers like nullability and comment.
        set_comment, set_nullable = None, None
        try:
            # loop over token blocks until either
            # - the next struct member definition starts at ,
            # - or the struct definition ends with ">"
            # - or a StopIteration is raised at the end of the string
            while iter.peek().value not in ",>":
                # GENERATED ALWAYS AS is simply ignored in this parsing
                if _ignore_generated(iter):
                    continue

                comment = _get_comment(iter)
                if comment is not None and not set_comment:
                    struct.fields[-1].metadata["comment"] = comment
                    set_comment = True
                    continue

                nullable = _get_nullable(iter)
                if nullable is not None and not set_nullable:
                    struct.fields[-1].nullable = nullable
                    set_nullable = True
                    continue

                # if neither of the above made us continue from this loop, get worried
                raise SchemaExtractionError(
                    f"unknown tokens after data type: {iter.peek()}"
                )
        except StopIteration:
            return struct

        # because of the last peek statement, we know that there must be a next value,
        # and it is one of "," or ">"
        final = next(iter).value
        if final == ">":
            # end of struct definition
            break
        elif final == ",":
            # on to the next struct member
            continue
        else:
            raise SchemaExtractionError("Logic error")

    return struct


def _get_nullable(iter) -> Optional[bool]:
    """See if a nullability follows. If it does, return the bool"""
    token = iter.peek()
    if str(token).upper() == "NULL":
        next(token)
        return True
    if re.match(r"NOT\s+NULL", str(token).upper()):
        next(iter)
        return False

    return None


def _get_comment(iter) -> Optional[str]:
    """See if a comment follows. If it does, return the string"""
    token = iter.peek()
    if token.value.upper() != "COMMENT":
        return None
    next(iter)

    comment = next(iter).value
    return comment.strip("'\"")


def _ignore_generated(iter):
    if iter.peek().value.upper() != "GENERATED":
        return False

    next(iter)

    if next(iter).value.upper() != "ALWAYS":
        raise SchemaExtractionError("Generated Always As expression is malformed")

    if next(iter).value.upper() != "AS":
        raise SchemaExtractionError("Generated Always As expression is malformed")

    if next(iter).value != "(":
        raise SchemaExtractionError("Generated Always As expression is malformed")

    _ignore_paren(iter, ")")
    return True


def _ignore_paren(iter, closing=")") -> None:
    for token in iter:
        if token.value == "(":
            _ignore_paren(iter, ")")
        if token.value == closing:
            return


def _get_data_type(iter) -> t.DataType:
    """Interpret the next word as a data type,
    then convert it to pyspark, consuming further tokens if necessary."""
    type_name_token = next(iter)
    type_name = type_name_token.value.upper()

    # Here below is a giant switch statement.
    # For each block, if it matches, it either returns or raises
    try:
        return dict(
            BOOLEAN=t.BooleanType(),
            BYTE=t.ByteType(),
            TINYINT=t.ByteType(),
            SHORT=t.ShortType(),
            SMALLINT=t.ShortType(),
            INT=t.IntegerType(),
            INTEGER=t.IntegerType(),
            LONG=t.LongType(),
            BIGINT=t.LongType(),
            FLOAT=t.FloatType(),
            REAL=t.FloatType(),
            DOUBLE=t.DoubleType(),
            DATE=t.DateType(),
            TIMESTAMP=t.TimestampType(),
            STRING=t.StringType(),
            BINARY=t.BinaryType(),
        )[type_name]
    except KeyError:
        pass

    if type_name in ["DECIMAL", "DEC", "NUMERIC"]:
        # we need to parse the (scale,precision) block if present
        paren_token = iter.peek()

        if paren_token.value != "(":
            return t.DecimalType()
        next(iter)

        precision = next(iter).value
        if next(iter).value != ",":
            raise SchemaExtractionError("expected comma after decimal scale")
        scale = next(iter).value
        if next(iter).value != ")":
            raise SchemaExtractionError("decimal precision section not properly closed")
        return t.DecimalType(precision=int(precision.strip()), scale=int(scale.strip()))

    if type_name == "INTERVAL":
        raise NotImplementedError()

    if type_name == "STRUCT":
        # we need to parse the struct members
        token = next(iter)
        if token.value != "<":
            raise SchemaExtractionError("Struct definition missing")
        return _get_schema(iter)

    if type_name == "ARRAY":
        token = next(iter)
        if token.value != "<":
            raise SchemaExtractionError("Array definition missing")
        element_type = _get_data_type(iter)
        token = next(iter)
        if token.value != ">":
            raise SchemaExtractionError("Array definition not closed as expected.")
        return t.ArrayType(elementType=element_type)

    if type_name == "MAP":
        token = next(iter)
        if token.value != "<":
            raise SchemaExtractionError("map definition missing")
        key_type = _get_data_type(iter)
        token = next(iter)
        if token.value != ",":
            raise SchemaExtractionError("map definition missing comma after key type")

        value_type = _get_data_type(iter)
        token = next(iter)
        if token.value != ">":
            raise SchemaExtractionError("map definition not closed as expected.")

        return t.MapType(key_type, value_type)

    raise NotImplementedError(f"Data type not implementd: {type_name}")
